# Copyright (C) 2014-2020 The BET Development Team

"""
This module contains functions for approximating jacobians of QoI maps.
All methods that cluster points around centers are written to return the
input_set._values in the following order : CENTERS, FOLLOWED BY THE CLUSTER
AROUND THE FIRST CENTER, THEN THE CLUSTER AROUND THE SECOND CENTER AND SO ON.
"""
import numpy as np
import scipy.spatial as spatial
import bet.sample as sample
import bet.util as util
import bet.sampling.LpGeneralizedSamples as lpsam


def sample_lp_ball(input_set, num_close, radius, p_num=2):
    r"""
    Pick num_close points in a the Lp ball of length 2*``radii_vec`` around a
    point in the input space, do this for each point in centers.  If this box
    extends outside of the domain of the input space, we sample the
    intersection.

    :param input_set: The input sample set.  Make sure the attribute
        ``_values`` is not ``None``
    :type input_set: :class:`~bet.sample.sample_set`
    :param int num_close: Number of points in each cluster
    :param radii_vec: Each side of the box will have length ``2*radii_vec[i]``
    :type radii_vec: :class:`numpy.ndarray` of shape (``input_dim``,)
    :param float p_num: :math:`0 < p \leq \infty`, p for the lp norm where
        infinity is ``numpy.inf``

    :rtype: :class:`~bet.sample.sample_set`
    :returns: Centers and clusters of samples near each center (values are
        :class:`numpy.ndarray` of shape ((``num_close+1``)*``num_centers``,
        ``input_dim``))

    """
    if input_set.get_values() is None:
        raise ValueError("You must have values to use this method.")
    input_dim = input_set.get_dim()
    centers = input_set.get_values()
    num_centers = input_set.check_num()
    input_domain = input_set.get_domain()

    cluster_set = sample.sample_set(input_dim)
    if input_domain is not None:
        cluster_set.set_domain(input_domain)
    cluster_set.set_values(centers)

    for i in range(num_centers):
        in_bounds = 0
        inflate = 1.0
        while in_bounds < num_close:
            # sample uniformly
            new_cluster = lpsam.Lp_generalized_uniform(input_dim,
                                                       num_close * inflate, p_num, radius, centers[i, :])
            # check bounds
            if input_domain is not None:
                cluster_set.update_bounds(num_close * inflate)
                left = np.all(np.greater_equal(new_cluster, cluster_set._left),
                              axis=1)
                right = np.all(np.less_equal(new_cluster, cluster_set._right),
                               axis=1)
                inside = np.logical_and(left, right)
                in_bounds = np.sum(inside)
                new_cluster = new_cluster[inside, :]
                # increase inflate
                inflate *= 10.0
            else:
                in_bounds = num_close

        if in_bounds > num_close:
            new_cluster = new_cluster[:num_close, :]
        cluster_set.append_values(new_cluster)

    # reset bounds
    cluster_set._left = None
    cluster_set._right = None
    cluster_set._width = None
    return cluster_set


def sample_linf_ball(input_set, num_close, radii_vec):
    r"""
    Pick num_close points in a the L-inifity ball of length 2*``radii_vec``
    around a point in the input space, do this for each point in centers.  If
    this box extends outside of the domain of the input space, we sample the
    intersection.

    :param input_set: The input sample set.  Make sure the attribute
        ``_values`` is not ``None``
    :type input_set: :class:`~bet.sample.sample_set`
    :param int num_close: Number of points in each cluster
    :param radii_vec: Each side of the box will have length ``2*radii_vec[i]``
    :type radii_vec: :class:`numpy.ndarray` of shape (``input_dim``,)

    :rtype: :class:`~bet.sample.sample_set`
    :returns: Centers and clusters of samples near each center (values are
        :class:`numpy.ndarray` of shape ((``num_close+1``)*``num_centers``,
        ``input_dim``))

    """
    return sample_lp_ball(input_set, num_close, radii_vec, p_num=np.inf)


def sample_l1_ball(input_set, num_close, radii_vec):
    r"""
    Pick num_close points in a the L-1 ball of length 2*``radii_vec`` around a
    point in the input space, do this for each point in centers.  If this box
    extends outside of the domain of the input space, we sample the
    intersection.

    :param input_set: The input sample set.  Make sure the attribute
        ``_values`` is not ``None``
    :type input_set: :class:`~bet.sample.sample_set`
    :param int num_close: Number of points in each cluster
    :param radii_vec: Each side of the box will have length ``2*radii_vec[i]``
    :type radii_vec: :class:`numpy.ndarray` of shape (``input_dim``,)

    :rtype: :class:`~bet.sample.sample_set`
    :returns: Centers and clusters of samples near each center (values are
        :class:`numpy.ndarray` of shape ((``num_close+1``)*``num_centers``,
        ``input_dim``))

    """
    return sample_lp_ball(input_set, num_close, radii_vec, p_num=1)


def pick_ffd_points(input_set, radii_vec):
    r"""
    Pick input_dim points, for each centers, for a forward finite
    difference gradient approximation.  The points are returned in the order:
    centers, followed by the cluster around the first center, then the cluster
    around the second center and so on.

    :param input_set: The input sample set.  Make sure the attribute _values is
        not None
    :type input_set: :class:`~bet.sample.sample_set`
    :param radii_vec: The radius of the stencil, along each axis
    :type radii_vec: :class:`numpy.ndarray` of shape (input_dim,)

    :rtype: :class:`~bet.sample.sample_set`
    :returns: Centers and clusters of samples near each center (values are
        :class:`numpy.ndarray` of shape ((``num_close+1``)*``num_centers``,
        ``input_dim``))

    """
    if input_set._values is None:
        raise ValueError("You must have values to use this method.")
    input_dim = input_set.get_dim()
    centers = input_set.get_values()
    num_centers = centers.shape[0]
    samples = np.repeat(centers, input_dim, axis=0)
    radii_vec = util.fix_dimensions_vector(radii_vec)

    # Construct a [num_centers*(input_dim+1), input_dim] matrix that
    # translates the centers to the FFD points.
    translate = np.tile(np.eye(input_dim) * radii_vec, (num_centers, 1))
    samples = samples + translate

    cluster_set = sample.sample_set(input_dim)
    if input_set.get_domain() is not None:
        cluster_set.set_domain(input_set.get_domain())
    cluster_set.set_values(centers)
    cluster_set.append_values(samples)
    return cluster_set


def pick_cfd_points(input_set, radii_vec):
    r"""
    Pick 2*input_dim points, for each center, for centered finite difference
    gradient approximation.  The center are not needed for the CFD gradient
    approximation, they are returned for consistency with the other methods and
    because of the common need to have not just the gradient but also the QoI
    value at the centers in adaptive sampling algorithms.The points are returned
    in the order: centers, followed by the cluster around the first center, then
    the cluster around the second center and so on.

    :param input_set: The input sample set.  Make sure the attribute _values is
        not None
    :type input_set: :class:`~bet.sample.sample_set`
    :param radii_vec: The radius of the stencil, along each axis
    :type radii_vec: :class:`numpy.ndarray` of shape (input_dim,)

    :rtype: :class:`~bet.sample.sample_set`
    :returns: Centers and clusters of samples near each center (values are
        :class:`numpy.ndarray` of shape ((``num_close+1``)*``num_centers``,
        ``input_dim``))

    """
    if input_set._values is None:
        raise ValueError("You must have values to use this method.")
    input_dim = input_set.get_dim()
    centers = input_set.get_values()
    num_centers = centers.shape[0]
    samples = np.repeat(centers, 2 * input_dim, axis=0)
    radii_vec = util.fix_dimensions_vector(radii_vec)

    # Contstruct a [num_centers*2*input_dim, input_dim] array that
    # translates the centers to the CFD points
    ident = np.eye(input_dim) * radii_vec
    translate = np.tile(np.append(ident, -ident, axis=0), (num_centers, 1))
    samples = samples + translate

    cluster_set = sample.sample_set(input_dim)
    if input_set.get_domain() is not None:
        cluster_set.set_domain(input_set.get_domain())
    cluster_set.set_values(centers)
    cluster_set.append_values(samples)
    return cluster_set


def radial_basis_function(r, kernel=None, ep=None):
    """
    Evaluate a chosen radial basis function.  Allow for the choice of several
    radial basis functions to use in
    :meth:~bet.sensitivity.gradients.calculate_gradients_rbf

    :param r: Distances from the reference point
    :type r: :class:`numpy.ndarray`
    :param string kernel: Choice of radial basis funtion. Default is C4Matern
    :param float ep: Shape parameter for the radial basis function.
        Default is 1.0

    :rtype: :class:`numpy.ndarray` of shape (r.shape)
    :returns: Radial basis function evaluated for each element of r

    """
    if ep is None:
        ep = 1.0

    if kernel is None or kernel is 'C4Matern':
        rbf = (1 + (ep * r) + (ep * r)**2 / 3) * np.exp(-ep * r)
    elif kernel is 'Gaussian':
        rbf = np.exp(-(ep * r)**2)
    elif kernel is 'Multiquadric':
        rbf = (1 + (ep * r)**2)**(0.5)
    elif kernel is 'InverseMultiquadric':
        rbf = 1 / ((1 + (ep * r)**2)**(0.5))
    else:
        raise ValueError("The kernel chosen is not currently available.")

    return rbf


def radial_basis_function_dxi(r, xi, kernel=None, ep=None):
    """
    Evaluate a partial derivative of a chosen radial basis function.  Allow for
    the choice of several radial basis functions to use in the
    :meth:~bet.sensitivity.gradients.calculate_gradients_rbf.

    :param r: Distances from the reference point
    :type r: :class:`numpy.ndarray`
    :param xi: Distances from the reference point in dimension i
    :type xi: :class:`numpy.ndarray`
    :param string kernel: Choice of radial basis funtion. Default is C4Matern
    :param float ep: Shape parameter for the radial basis function.
        Default is 1.0

    :rtype: :class:`numpy.ndarray` of shape (r.shape)
    :returns: Radial basis function evaluated for each element of r

    """
    if ep is None:
        ep = 1.0

    if kernel is None or kernel is 'C4Matern':
        rbfdxi = -(ep**2 * xi * np.exp(-ep * r) * (ep * r + 1)) / 3
    elif kernel is 'Gaussian':
        rbfdxi = -2 * ep**2 * xi * np.exp(-(ep * r)**2)
    elif kernel is 'Multiquadric':
        rbfdxi = (ep**2 * xi) / ((1 + (ep * r)**2)**(0.5))
    elif kernel is 'InverseMultiquadric':
        rbfdxi = -(ep**2 * xi) / ((1 + (ep * r)**2)**(1.5))
    else:
        raise ValueError("The kernel chosen is not currently available")

    return rbfdxi


def calculate_gradients_rbf(cluster_discretization, num_centers=None,
                            num_neighbors=None, RBF=None, ep=None, normalize=True):
    r"""
    Approximate gradient vectors at ``num_centers, centers.shape[0]`` points
    in the parameter space for each QoI map using a radial basis function
    interpolation method.

    :param cluster_discretization: Must contain input and output values for the
        sample clusters.
    :type cluster_discretization: :class:`~bet.sample.discretization`
    :param int num_centers: The number of cluster centers.
    :param int num_neighbors: Number of nearest neighbors to use in gradient
        approximation. Default value is ``input_dim + 2``
    :param string RBF: Choice of radial basis function. Default is Gaussian
    :param float ep: Choice of shape parameter for radial basis function.
        Default value is 1.0
    :param boolean normalize:  If normalize is True, normalize each gradient
        vector

    :rtype: :class:`~bet.sample.discretization`
    :returns: A new :class:`~bet.sample.discretization` that contains only the
        centers of the clusters and their associated ``_jacobians`` which are
        tensor representation of the gradient vectors of each QoI map at each
        point in centers :class:`numpy.ndarray` of shape (num_samples,
        output_dim, input_dim)

    """
    if cluster_discretization._input_sample_set.get_values() is None \
            or cluster_discretization._output_sample_set.get_values() is None:
        raise ValueError("You must have values to use this method.")
    samples = cluster_discretization._input_sample_set.get_values()
    data = cluster_discretization._output_sample_set.get_values()

    input_dim = cluster_discretization._input_sample_set.get_dim()
    num_model_samples = cluster_discretization.check_nums()
    output_dim = cluster_discretization._output_sample_set.get_dim()

    if num_neighbors is None:
        num_neighbors = input_dim + 2
    if ep is None:
        ep = 1.0
    if RBF is None:
        RBF = 'Gaussian'

    # If centers is None we assume the user chose clusters of size
    # input_dim + 2
    if num_centers is None:
        num_centers = num_model_samples // (input_dim + 2)
    centers = samples[:num_centers, :]

    rbf_tensor = np.zeros([num_centers, num_model_samples, input_dim])
    gradient_tensor = np.zeros([num_centers, output_dim, input_dim])

    # For each center, interpolate the data using the rbf chosen and
    # then evaluate the partial derivative of that interpolant at the desired
    # point.
    for c in range(num_centers):
        # Find the k nearest neighbors and their distances to centers[c,:]
        [r, nearest] = cluster_discretization._input_sample_set.query(
            centers[c, :], k=num_neighbors)
        r = np.tile(r, (input_dim, 1))

        # Compute the linf distances to each of the nearest neighbors
        diffVec = (centers[c, :] - samples[nearest, :]).transpose()

        # Compute the l2 distances between pairs of nearest neighbors
        distMat = spatial.distance_matrix(
            samples[nearest, :], samples[nearest, :])

        # Solve for the rbf weights using interpolation conditions and
        # evaluate the partial derivatives
        rbf_mat_values = \
            np.linalg.solve(radial_basis_function(distMat, RBF),
                            radial_basis_function_dxi(r, diffVec, RBF, ep)
                            .transpose()).transpose()

        # Construct the finite difference matrices
        rbf_tensor[c, nearest, :] = rbf_mat_values.transpose()

    gradient_tensor = rbf_tensor.transpose(
        2, 0, 1).dot(data).transpose(1, 2, 0)

    if normalize:
        # Compute the norm of each vector
        norm_gradient_tensor = np.linalg.norm(gradient_tensor, ord=1, axis=2)

        # If it is a zero vector (has 0 norm), set norm=1, avoid divide by zero
        norm_gradient_tensor[norm_gradient_tensor == 0] = 1.0

        # Normalize each gradient vector
        gradient_tensor = gradient_tensor / np.tile(norm_gradient_tensor,
                                                    (input_dim, 1, 1)).transpose(1, 2, 0)

    center_input_sample_set = sample.sample_set(input_dim)
    center_input_sample_set.set_values(samples[:num_centers, :])
    if cluster_discretization._input_sample_set.get_domain() is not None:
        center_input_sample_set.set_domain(cluster_discretization.
                                           _input_sample_set.get_domain())
    center_input_sample_set.set_jacobians(gradient_tensor)
    center_output_sample_set = sample.sample_set(output_dim)
    center_output_sample_set.set_values(data[:num_centers, :])
    if cluster_discretization._output_sample_set.get_domain() is not None:
        center_output_sample_set.set_domain(cluster_discretization.
                                            _output_sample_set.get_domain())
    # center_output_sample_set.set_jacobians(gradient_tensor.transpose())
    center_discretization = sample.discretization(center_input_sample_set,
                                                  center_output_sample_set)
    return center_discretization


def calculate_gradients_ffd(cluster_discretization, normalize=True):
    """
    Approximate gradient vectors at ``num_centers, centers.shape[0]`` points
    in the parameter space for each QoI map.  THIS METHOD IS DEPENDENT ON USING
    :meth:~bet.sensitivity.gradients.pick_ffd_points TO CHOOSE SAMPLES FOR THE
    FFD STENCIL AROUND EACH CENTER. THE ORDERING MATTERS.

    :param cluster_discretization: Must contain input and output values for the
        sample clusters.
    :type cluster_discretization: :class:`~bet.sample.discretization`
    :param boolean normalize:  If normalize is True, normalize each gradient
        vector

    :rtype: :class:`~bet.sample.discretization`
    :returns: A new :class:`~bet.sample.discretization` that contains only the
        centers of the clusters and their associated ``_jacobians`` which are
        tensor representation of the gradient vectors of each QoI map at each
        point in centers :class:`numpy.ndarray` of shape (num_samples,
        output_dim, input_dim)

    """
    if cluster_discretization._input_sample_set.get_values() is None \
            or cluster_discretization._output_sample_set.get_values() is None:
        raise ValueError("You must have values to use this method.")
    samples = cluster_discretization._input_sample_set.get_values()
    data = cluster_discretization._output_sample_set.get_values()

    input_dim = cluster_discretization._input_sample_set.get_dim()
    num_model_samples = cluster_discretization.check_nums()
    output_dim = cluster_discretization._output_sample_set.get_dim()

    num_model_samples = cluster_discretization.check_nums()
    input_dim = cluster_discretization._input_sample_set.get_dim()
    num_centers = num_model_samples // (input_dim + 1)

    # Find radii_vec from the first cluster of samples
    radii_vec = samples[num_centers:num_centers + input_dim, :] - samples[0, :]
    radii_vec = util.fix_dimensions_vector_2darray(radii_vec.diagonal())

    # Clean the data
    data = util.clean_data(data)
    gradient_tensor = np.zeros([num_centers, output_dim, input_dim])

    radii_vec = np.tile(np.repeat(radii_vec, output_dim, axis=1), [num_centers,
                                                                   1])

    # Compute the gradient vectors using the standard FFD stencil
    gradient_mat = (data[num_centers:] - np.repeat(data[0:num_centers],
                                                   input_dim, axis=0)) * (1. / radii_vec)

    # Reshape and organize
    gradient_tensor = np.reshape(gradient_mat.transpose(), [output_dim,
                                                            input_dim, num_centers], order='F').transpose(2, 0, 1)

    if normalize:
        # Compute the norm of each vector
        norm_gradient_tensor = np.linalg.norm(gradient_tensor, ord=1, axis=2)

        # If it is a zero vector (has 0 norm), set norm=1, avoid divide by zero
        norm_gradient_tensor[norm_gradient_tensor == 0] = 1.0

        # Normalize each gradient vector
        gradient_tensor = gradient_tensor / np.tile(norm_gradient_tensor,
                                                    (input_dim, 1, 1)).transpose(1, 2, 0)

    center_input_sample_set = sample.sample_set(input_dim)
    center_input_sample_set.set_values(samples[:num_centers, :])
    if cluster_discretization._input_sample_set.get_domain() is not None:
        center_input_sample_set.set_domain(cluster_discretization.
                                           _input_sample_set.get_domain())
    center_input_sample_set.set_jacobians(gradient_tensor)
    center_output_sample_set = sample.sample_set(output_dim)
    center_output_sample_set.set_values(data[:num_centers, :])
    if cluster_discretization._output_sample_set.get_domain() is not None:
        center_output_sample_set.set_domain(cluster_discretization.
                                            _output_sample_set.get_domain())
    # center_output_sample_set.set_jacobians(gradient_tensor.transpose())
    center_discretization = sample.discretization(center_input_sample_set,
                                                  center_output_sample_set)
    return center_discretization


def calculate_gradients_cfd(cluster_discretization, normalize=True):
    """
    Approximate gradient vectors at ``num_centers, centers.shape[0]`` points
    in the parameter space for each QoI map.  THIS METHOD IS DEPENDENT
    ON USING :meth:~bet.sensitivity.pick_cfd_points TO CHOOSE SAMPLES FOR THE
    CFD STENCIL AROUND EACH CENTER.  THE ORDERING MATTERS.

    :param cluster_discretization: Must contain input and output values for the
        sample clusters.
    :type cluster_discretization: :class:`~bet.sample.discretization`
    :param boolean normalize:  If normalize is True, normalize each gradient
        vector

    :rtype: :class:`~bet.sample.discretization`
    :returns: A new :class:`~bet.sample.discretization` that contains only the
        centers of the clusters and their associated ``_jacobians`` which are
        tensor representation of the gradient vectors of each QoI map at each
        point in centers :class:`numpy.ndarray` of shape (num_samples,
        output_dim, input_dim)

    """
    if cluster_discretization._input_sample_set.get_values() is None \
            or cluster_discretization._output_sample_set.get_values() is None:
        raise ValueError("You must have values to use this method.")
    samples = cluster_discretization._input_sample_set.get_values()
    data = cluster_discretization._output_sample_set.get_values()

    input_dim = cluster_discretization._input_sample_set.get_dim()
    num_model_samples = cluster_discretization.check_nums()
    output_dim = cluster_discretization._output_sample_set.get_dim()

    num_model_samples = cluster_discretization.check_nums()
    input_dim = cluster_discretization._input_sample_set.get_dim()

    num_centers = num_model_samples // (2 * input_dim + 1)

    # Find radii_vec from the first cluster of samples
    radii_vec = samples[num_centers:num_centers + input_dim, :] - samples[0, :]
    radii_vec = util.fix_dimensions_vector_2darray(radii_vec.diagonal())

    # Clean the data
    data = util.clean_data(data[num_centers:])
    gradient_tensor = np.zeros([num_centers, output_dim, input_dim])

    radii_vec = np.tile(np.repeat(radii_vec, output_dim, axis=1), [num_centers,
                                                                   1])

    # Construct indices for CFD gradient approxiation
    inds = np.repeat(np.arange(0, 2 * input_dim * num_centers, 2 * input_dim),
                     input_dim) + np.tile(np.arange(0, input_dim), num_centers)
    inds = np.array([inds, inds + input_dim]).transpose()

    gradient_mat = (data[inds[:, 0]] - data[inds[:, 1]]) * (0.5 / radii_vec)

    # Reshape and organize
    gradient_tensor = np.reshape(gradient_mat.transpose(), [output_dim,
                                                            input_dim, num_centers], order='F').transpose(2, 0, 1)

    if normalize:
        # Compute the norm of each vector
        norm_gradient_tensor = np.linalg.norm(gradient_tensor, ord=1, axis=2)

        # If it is a zero vector (has 0 norm), set norm=1, avoid divide by zero
        norm_gradient_tensor[norm_gradient_tensor == 0] = 1.0

        # Normalize each gradient vector
        gradient_tensor = gradient_tensor / np.tile(norm_gradient_tensor,
                                                    (input_dim, 1, 1)).transpose(1, 2, 0)

    center_input_sample_set = sample.sample_set(input_dim)
    center_input_sample_set.set_values(samples[:num_centers, :])
    if cluster_discretization._input_sample_set.get_domain() is not None:
        center_input_sample_set.set_domain(cluster_discretization.
                                           _input_sample_set.get_domain())
    center_input_sample_set.set_jacobians(gradient_tensor)
    center_output_sample_set = sample.sample_set(output_dim)
    center_output_sample_set.set_values(data[:num_centers, :])
    if cluster_discretization._output_sample_set.get_domain() is not None:
        center_output_sample_set.set_domain(cluster_discretization.
                                            _output_sample_set.get_domain())
    # center_output_sample_set.set_jacobians(gradient_tensor.transpose())
    center_discretization = sample.discretization(center_input_sample_set,
                                                  center_output_sample_set)
    return center_discretization



# Copyright (C) 2014-2020 The BET Development Team

"""
This module provides methods used to plot two-dimensional domains and/or
two-dimensional slices/projections of domains.

"""

import os
from itertools import combinations
import numpy as np
import matplotlib.tri as tri
import matplotlib.pyplot as plt
# plt.rc('text', usetex=True)
# plt.rc('font', family='serif')
from matplotlib.lines import Line2D
from mpl_toolkits.mplot3d import Axes3D
import bet.sample as sample
import bet.util as util

markers = []
for m in Line2D.markers:
    try:
        if len(m) == 1 and m != ' ':
            markers.append(m)
    except TypeError:
        pass

colors = ('b', 'g', 'r', 'c', 'm', 'y', 'k')


class dim_not_matching(Exception):
    """
    Exception for when the dimension is inconsistent.
    """


class bad_object(Exception):
    """
    Exception for when the wrong type of object is used.
    """


def scatter_2D(sample_obj, sample_nos=None, color=None, ref_sample=None,
               save=True, interactive=False, xlabel=r'$x$', ylabel=r'$y$', cbar_label=None,
               filename='scatter2d', file_extension=".png", markersize=75):
    r"""
    Creates a two-dimensional scatter plot of the samples within the sample
    object colored by ``color`` (usually an array of pointwise probability
    density values).  A reference sample (``ref_sample``) can be chosen by the
    user.  This reference sample will be plotted as a mauve circle twice the
    size of the other markers.

    .. note::

        Do not specify the file extension in BOTH ``filename`` and
        ``file_extension``.

    :param sample_obj: contains samples to create scatter plot
    :type sample_obj: :class:`~bet.sample.sample_set_base`
    :param list sample_nos: indicies of the samples to plot
    :param color: values to color the samples by
    :type color: :class:`numpy.ndarray`
    :param ref_sample: reference parameter value
    :type ref_sample: :class:`numpy.ndarray` of shape (ndim,)
    :param bool save: flag whether or not to save the figure
    :param bool interactive: flag whether or not to show the figure
    :param string xlabel: x-axis label
    :param string ylabel: y-axis label
    :param string cbar_label: color bar label
    :param string filename: filename to save the figure as
    :param string file_extension: file extension

    """
    if not isinstance(sample_obj, sample.sample_set_base):
        raise bad_object("Improper sample object")
    # check dimension of data to plot
    if sample_obj.get_dim() != 2:
        raise dim_not_matching("Cannot create 2D plot of non-2D sample "
                               "object")
    if ref_sample is None:
        ref_sample = sample_obj._reference_value

    # plot all of the samples by default
    if sample_nos is None:
        sample_nos = np.arange(sample_obj.get_values().shape[0])
    # color all of the samples uniformly by default and set the default
    # to the default colormap of matplotlib
    if color is None:
        color = np.ones((sample_obj.get_values().shape[0],))
        cbar_label = "index"
        cmap = None
    else:
        cmap = plt.cm.PuBu

    if cbar_label is None:
        cbar_label = r'$\rho_\mathcal{D}(q)$'

    color = color[sample_nos]
    # handle scalar scatterplots in in matplotlib>=3.0
    color = np.reshape(color, -1)

    fig = plt.figure(constrained_layout=True)
    # create the scatter plot for the samples specified by sample_nos
    plt.scatter(np.reshape(sample_obj.get_values()[sample_nos, 0], -1),
                np.reshape(sample_obj.get_values()[sample_nos, 1], -1),
                c=color, s=markersize, alpha=.75, linewidth=.1, cmap=cmap)
    # add a colorbar and label for the colorbar usually we just assume the
    # samples are colored by the pointwise probability density on the data
    # space
    cbar = plt.colorbar()

    cbar.set_label(cbar_label)

    # if there is a reference value plot it with a notiable marker
    if ref_sample is not None:
        plt.scatter(ref_sample[0], ref_sample[1], c='m', s=2 * markersize)
    if save:
        plt.autoscale(tight=True)
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        if "." not in filename:
            full_filename = filename + file_extension
        else:
            full_filename = filename
        plt.savefig(full_filename, bbox_inches='tight', transparent=True,
                    pad_inches=0)
    if interactive:
        plt.show()
    else:
        plt.close()


def scatter_2D_input(my_disc, sample_nos=None, color=None, ref_sample=None,
                     save=True, interactive=False, xlabel=r'$\lambda_1$', ylabel=r'$\lambda_2$', cbar_label=None,
                     filename='scatter2d_input', file_extension=".png", markersize=75):
    r"""
    Creates a two-dimensional scatter plot of the input samples within the
    discretization object colored by ``color`` (usually an array of pointwise
    probability density values).  A reference sample (``ref_sample``) can be
    chosen by the user.  This reference sample will be plotted as a mauve
    circle twice the size of the other markers.

    .. note::

        Do not specify the file extension in BOTH ``filename`` and
        ``file_extension``.

    :param my_disc: contains samples (`my_disc._input_sample_set``) to create
        scatter plot
    :type my_disc: :class:`~bet.sample.discretization`
    :param list sample_nos: indicies of the samples to plot
    :param color: values to color the samples by
    :type color: :class:`numpy.ndarray` or string (volumes, probabilities,
        radii, normalized radii, or error id)
    :param ref_sample: reference parameter value
    :type ref_sample: :class:`numpy.ndarray` of shape (ndim,)
    :param bool save: flag whether or not to save the figure
    :param bool interactive: flag whether or not to show the figure
    :param string xlabel: x-axis label
    :param string ylabel: y-axis label
    :param string cbar_label: color bar label
    :param string filename: filename to save the figure as
    :param string file_extension: file extension

    """
    if not isinstance(my_disc, sample.discretization):
        raise bad_object("Improper discretization object")

    sample_obj = my_disc.get_input_sample_set()

    if color is "volumes":
        cbar_label = r'$\mu_\Lambda(\mathcal{V}_{i,N})$'
        color = sample_obj.get_volumes()
    elif color is "probabilities":
        cbar_label = r'$\P_\Lambda(\mathcal{V}_{i,N})$'
        color = sample_obj.get_probabilities()
    elif color is "radii":
        cbar_label = r'$diam_\Lambda(\mathcal{V}_{i,N})$'
        color = sample_obj._radii
    elif color is "normalized radii":
        cbar_label = r'$diam_\Lambda_{norm}(\mathcal{V}_{i,N})$'
        color = sample_obj._normalized_radii
    elif color is "error_id":
        cbar_label = 'error id'
        color = sample_obj.get_error_id()

    scatter_2D(sample_obj, sample_nos, color, ref_sample,
               save, interactive, xlabel, ylabel, cbar_label, filename,
               file_extension, markersize)


def scatter_2D_output(my_disc, sample_nos=None, color=None, ref_sample=None,
                      save=True, interactive=False, xlabel=r'$q_1$', ylabel=r'$q_2$', cbar_label=None,
                      filename='scatter2d_output', file_extension=".png", markersize=75):
    r"""
    Creates a two-dimensional scatter plot of the output samples within the
    discretization object colored by ``color`` (usually an array of pointwise
    probability density values).  A reference sample (``ref_sample``) can be
    chosen by the user.  This reference sample will be plotted as a mauve
    circle twice the size of the other markers.

    .. note::

        Do not specify the file extension in BOTH ``filename`` and
        ``file_extension``.

    :param my_disc: contains samples (`my_disc._output_sample_set``) to create
        scatter plot
    :type my_disc: :class:`~bet.sample.discretization`
    :param list sample_nos: indicies of the samples to plot
    :param color: values to color the samples by
    :type color: :class:`numpy.ndarray` or string (volumes, probabilities,
        radii, normalized radii, or error id)
    :param ref_sample: reference parameter value
    :type ref_sample: :class:`numpy.ndarray` of shape (ndim,)
    :param bool save: flag whether or not to save the figure
    :param bool interactive: flag whether or not to show the figure
    :param string xlabel: x-axis label
    :param string ylabel: y-axis label
    :param string cbar_label: color bar label
    :param string filename: filename to save the figure as
    :param string file_extension: file extension

    """
    if not isinstance(my_disc, sample.discretization):
        raise bad_object("Improper discretization object")

    sample_obj = my_disc.get_output_sample_set()

    if color is "volumes":
        cbar_label = r'$\mu_\mathcal{D}(\mathcal{I}_{j,M})$'
        color = sample_obj.get_volumes()
    elif color is "probabilities":
        cbar_label = r'$\P_\mathcal{D}(\mathcal{I}_{j,M})$'
        color = sample_obj.get_probabilities()
    elif color is "radii":
        cbar_label = r'$diam_\mathcal{D}(\mathcal{I}_{j,M})$'
        color = sample_obj._radii
    elif color is "normalized radii":
        cbar_label = r'$diam_\mathcal{D}_{norm}(\mathcal{I}_{j,M})$'
        color = sample_obj._normalized_radii
    elif color is "error_id":
        cbar_label = 'error id'
        color = sample_obj.get_error_id()

    scatter_2D(sample_obj, sample_nos, color, ref_sample,
               save, interactive, xlabel, ylabel, cbar_label, filename,
               file_extension, markersize)


def scatter_3D(sample_obj, sample_nos=None, color=None, ref_sample=None,
               save=True, interactive=False, xlabel='x', ylabel='y', zlabel='z',
               cbar_label=None, filename="scatter3d", file_extension=".png", markersize=75):
    r"""
    Creates a three-dimensional scatter plot of samples within the sample
    object colored by ``color`` (usually an array of pointwise probability
    density values). A reference sample (``ref_sample``) can be chosen by the
    user.  This reference sample will be plotted as a mauve circle twice the
    size of the other markers.

    .. note::

        Do not specify the file extension in BOTH ``filename`` and
        ``file_extension``.

    :param sample_obj: Object containing the samples to plot
    :type sample_obj: :class:`~bet.sample.sample_set_base`
    :param list sample_nos: indicies of the samples to plot
    :param color: values to color the samples by
    :type color: :class:`numpy.ndarray`
    :param ref_sample: reference parameter value
    :type ref_sample: :class:`numpy.ndarray` of shape (ndim,)
    :param bool save: flag whether or not to save the figure
    :param bool interactive: flag whether or not to show the figure
    :param string xlabel: x-axis label
    :param string ylabel: y-axis label
    :param string zlabel: z-axis label
    :param string cbar_label: color bar label
    :param string filename: filename to save the figure as
    :param string file_extension: file extension

    """
    if not isinstance(sample_obj, sample.sample_set_base):
        raise bad_object("Improper sample object")
    # check dimension of data to plot
    if sample_obj.get_dim() != 3:
        raise dim_not_matching("Cannot create 3D plot of non-3D sample "
                               "object")

    if ref_sample is None:
        ref_sample = sample_obj._reference_value

    # plot all of the samples by default
    if sample_nos is None:
        sample_nos = np.arange(sample_obj.get_values().shape[0])
    # color all of the samples uniformly by default and set the default
    # to the default colormap of matplotlib
    if color is None:
        color = np.ones((sample_obj.get_values().shape[0],))
        cmap = None
    else:
        cmap = plt.cm.PuBu
    color = np.reshape(color[sample_nos], -1)
    # create the scatter plot for the samples specified by sample_nos
    fig = plt.figure(constrained_layout=True)
    ax = fig.add_subplot(111, projection='3d')
    p = ax.scatter(np.reshape(sample_obj.get_values()[sample_nos, 0], -1),
                   np.reshape(sample_obj.get_values()[sample_nos, 1], -1),
                   np.reshape(sample_obj.get_values()[sample_nos, 2], -1),
                   alpha=.75, linewidth=.1, c=color, s=markersize, cmap=cmap)
    # add a colorbar and label for the colorbar usually we just assume the
    # samples are colored by the pointwise probability density on the data
    # space
    cbar = fig.colorbar(p)
    if cbar_label is None:
        cbar_label = r'$\rho_\mathcal{D}(q)$'
    cbar.set_label(cbar_label)
    # if there is a reference value plot it with a notiable marker
    if ref_sample is not None:
        ax.scatter(ref_sample[0], ref_sample[1], ref_sample[2], c='m', s=2 *
                   markersize)
    ax.autoscale(tight=True)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_zlabel(zlabel)
    if save:
        if "." not in filename:
            full_filename = filename + file_extension
        else:
            full_filename = filename
        plt.savefig(full_filename, bbox_inches='tight', transparent=True,
                    pad_inches=0)
    if interactive:
        plt.show()
    else:
        plt.close()


def scatter_3D_input(my_disc, sample_nos=None, color=None, ref_sample=None,
                     save=True, interactive=False,
                     xlabel=r'$\lambda_1$', ylabel=r'$\lambda_2$', zlabel=r'$\lambda_3$',
                     cbar_label=None, filename="scatter3d_input", file_extension=".png", markersize=75):
    r"""
    Creates a three-dimensional scatter plot of input samples within the
    discretization object colored by ``color`` (usually an array of pointwise
    probability density values). A reference sample (``ref_sample``) can be
    chosen by the user.  This reference sample will be plotted as a mauve
    circle twice the size of the other markers.

    .. note::

        Do not specify the file extension in BOTH ``filename`` and
        sample_obj = my_disc.get_input_sample_set() ``file_extension``.

    :param my_disc: contains samples (`my_disc._output_sample_set``) to create
        scatter plot
    :type my_disc: :class:`~bet.sample.discretization`
    :param list sample_nos: indicies of the samples to plot
    :param color: values to color the samples by
    :type color: :class:`numpy.ndarray` or string (volumes, probabilities,
        radii, normalized radii, or error id)
    :param ref_sample: reference parameter value
    :type ref_sample: :class:`numpy.ndarray` of shape (ndim,)
    :param bool save: flag whether or not to save the figure
    :param bool interactive: flag whether or not to show the figure
    :param string xlabel: x-axis label
    :param string ylabel: y-axis label
    :param string zlabel: z-axis label
    :param string cbar_label: color bar label
    :param string filename: filename to save the figure as
    :param string file_extension: file extension

    """
    if not isinstance(my_disc, sample.discretization):
        raise bad_object("Improper discretization object")

    sample_obj = my_disc.get_input_sample_set()

    if color is "volumes":
        cbar_label = r'$\mu_\Lambda(\mathcal{V}_{i,N})$'
        color = sample_obj.get_volumes()
    elif color is "probabilities":
        cbar_label = r'$\P_\Lambda(\mathcal{V}_{i,N})$'
        color = sample_obj.get_probabilities()
    elif color is "radii":
        cbar_label = r'$diam_\Lambda(\mathcal{V}_{i,N})$'
        color = sample_obj._radii
    elif color is "normalized radii":
        cbar_label = r'$diam_\Lambda_{norm}(\mathcal{V}_{i,N})$'
        color = sample_obj._normalized_radii
    elif color is "error_id":
        cbar_label = 'error id'
        color = sample_obj.get_error_id()

    scatter_3D_input(sample_obj, sample_nos, color,
                     ref_sample, save, interactive, xlabel, ylabel, zlabel, cbar_label,
                     filename, file_extension, markersize)


def scatter_3D_output(my_disc, sample_nos=None, color=None, ref_sample=None,
                      save=True, interactive=False,
                      xlabel=r'$q_1$', ylabel=r'$q_2$', zlabel=r'$q_3$',
                      cbar_label=None, filename="scatter3d_output", file_extension=".png", markersize=75):
    r"""
    Creates a three-dimensional scatter plot of output samples within the
    discretization object colored by ``color`` (usually an array of pointwise
    probability density values). A reference sample (``ref_sample``) can be
    chosen by the user.  This reference sample will be plotted as a mauve
    circle twice the size of the other markers.

    .. note::

        Do not specify the file extension in BOTH ``filename`` and
        ``file_extension``.

    :param my_disc: contains samples (`my_disc._output_sample_set``) to create
        scatter plot
    :type my_disc: :class:`~bet.sample.discretization`
    :param list sample_nos: indicies of the samples to plot
    :param color: values to color the samples by
    :type color: :class:`numpy.ndarray` or string (volumes, probabilities,
        radii, normalized radii, or error id)
    :param ref_sample: reference parameter value
    :type ref_sample: :class:`numpy.ndarray` of shape (ndim,)
    :param bool save: flag whether or not to save the figure
    :param bool interactive: flag whether or not to show the figure
    :param string xlabel: x-axis label
    :param string ylabel: y-axis label
    :param string zlabel: z-axis label
    :param string cbar_label: color bar label
    :param string filename: filename to save the figure as
    :param string file_extension: file extension

    """
    if not isinstance(my_disc, sample.discretization):
        raise bad_object("Improper discretization object")

    sample_obj = my_disc.get_output_sample_set()

    if color is "volumes":
        cbar_label = r'$\mu_\mathcal{D}(\mathcal{I}_{j,M})$'
        color = sample_obj.get_volumes()
    elif color is "probabilities":
        cbar_label = r'$\P_\mathcal{D}(\mathcal{I}_{j,M})$'
        color = sample_obj.get_probabilities()
    elif color is "radii":
        cbar_label = r'$diam_\mathcal{D}(\mathcal{I}_{j,M})$'
        color = sample_obj._radii
    elif color is "normalized radii":
        cbar_label = r'$diam_\mathcal{D}_{norm}(\mathcal{I}_{j,M})$'
        color = sample_obj._normalized_radii
    elif color is "error_id":
        cbar_label = 'error id'
        color = sample_obj.get_error_id()

    scatter_3D_output(sample_obj, sample_nos, color,
                      ref_sample, save, interactive, xlabel, ylabel, zlabel, cbar_label,
                      filename, file_extension, markersize)


def scatter_rhoD(sample_obj, ref_sample=None, sample_nos=None, io_flag='input',
                 rho_D=None, dim_nums=None, label_char=None, showdim=None, save=True,
                 interactive=False, file_extension=".png", markersize=75):
    r"""
    Create scatter plots of samples within the sample object colored by
    ``color`` (usually an array of pointwise probability density values).  A
    reference sample (``ref_sample``) can be chosen by the user.  This reference
    sample will be plotted as a mauve circle twice the size of the other
    markers.

    .. note::

        Do not specify the file extension in BOTH ``filename`` and
        ``file_extension``.

    :param sample_obj: Object containing the samples to plot
    :type sample_obj: :class:`~bet.sample.discretization`
        or :class:`~bet.sample.sample_set_base`
    :param ref_sample: reference parameter value
    :type ref_sample: :class:`numpy.ndarray` of shape (ndim,)
    :param list sample_nos: sample numbers to plot
    :param string io_flag: Either `input` or `output`. If ``sample_obj`` is a
        :class:`~bet.sample.discretization` object flag whether or not put plot
        input or output.
    :param rho_D: probability density function on D
    :type rho_D: callable function that takes a :class:`np.array` and returns a
        :class:`numpy.ndarray`
    :param list dim_nums: integers representing domain coordinate
        numbers to plot (e.g. i, where :math:`\x_i` is a coordinate in the
        input/output space).
    :param string label_char: character to use to label coordinate axes
    :param int showdim: 2 or 3, flag to determine whether or not to show
        pairwise or tripletwise parameter sample scatter plots in 2 or 3
        dimensions
    :param bool save: flag whether or not to save the figure
    :param bool interactive: flag whether or not to show the figure
    :param string file_extension: file extension

    """
    # If there is density function given determine the pointwise probability
    # values of each sample based on the value in the data space. Otherwise,
    # color the samples in numerical order.
    rD = None
    if isinstance(sample_obj, sample.discretization):
        if rho_D is not None:
            rD = rho_D(sample_obj._output_sample_set.get_values())
        if io_flag == 'input':
            sample_obj = sample_obj._input_sample_set
        else:
            sample_obj = sample_obj._output_sample_set
    elif isinstance(sample_obj, sample.sample_set_base):
        if io_flag == 'output':
            rD = rho_D(sample_obj.get_values())
    else:
        raise bad_object("Improper sample object")

    if ref_sample is None:
        ref_sample = sample_obj._reference_value

    if rD is None:
        rD = np.ones(sample_obj.get_values().shape[0])

    if label_char is None:
        if io_flag == 'input':
            label_char = r'$\lambda_'
            prefix = 'input_'
        elif io_flag == 'output':
            label_char = r'$q_'
            prefix = 'output_'
        else:  # no io_flag
            label_char = r'$x_'
            prefix = 'rhoD_'
    else:
        label_char = r'$x_'
        prefix = 'rhoD_'

    # If no specific coordinate numbers are given for the parameter coordinates
    # (e.g. i, where \lambda_i is a coordinate in the parameter space), then
    # set them to be the the counting numbers.
    if dim_nums is None:
        dim_nums = 1 + np.array(np.arange(sample_obj.get_values().shape[1]))
    # Create the labels based on the user selected parameter coordinates
    xlabel = label_char + r'{' + str(dim_nums[0]) + '}$'
    ylabel = label_char + r'{' + str(dim_nums[1]) + '}$'
    savename = prefix + 'samples_cs'
    # Plot 2 or 3 dimensional scatter plots of the samples colored by rD.
    if sample_obj.get_dim() == 2:
        scatter_2D(sample_obj, sample_nos, rD, ref_sample, save,
                   interactive, xlabel, ylabel, None, savename,
                   markersize=markersize)
    elif sample_obj.get_dim() > 2 and showdim == 2:
        temp_obj = sample.sample_set(2)
        for x, y in combinations(dim_nums, 2):
            xlabel = label_char + r'{' + str(x) + '}$'
            ylabel = label_char + r'{' + str(y) + '}$'
            savename = prefix + 'samples_x' + str(x) + 'x' + str(y) + '_cs'
            temp_obj.set_values(sample_obj.get_values()[:, [x - 1, y - 1]])
            if ref_sample is not None:
                scatter_2D(temp_obj, sample_nos, rD, ref_sample[[x - 1, y - 1]], save,
                           interactive, xlabel, ylabel, None, savename,
                           markersize=markersize)
            else:
                scatter_2D(temp_obj, sample_nos, rD, ref_sample, save,
                           interactive, xlabel, ylabel, None, savename,
                           markersize=markersize)
    elif sample_obj.get_dim() == 3:
        zlabel = label_char + r'{' + str(dim_nums[2]) + '}$'
        scatter_3D(sample_obj, sample_nos, rD, ref_sample, save,
                   interactive, xlabel, ylabel, zlabel, None, savename,
                   markersize=markersize)
    elif sample_obj.get_dim() > 3 and showdim == 3:
        temp_obj = sample.sample_set(3)
        for x, y, z in combinations(dim_nums, 3):
            xlabel = label_char + r'{' + str(x) + '}$'
            ylabel = label_char + r'{' + str(y) + '}$'
            zlabel = label_char + r'{' + str(z) + '}$'
            savename = prefix + 'samples_x' + str(x) + 'x' + str(y) + 'x' +\
                str(z) + '_cs'
            temp_obj.set_values(sample_obj.get_values()[:, [x - 1, y - 1,
                                                            z - 1]])
            if ref_sample is not None:
                scatter_3D(temp_obj, sample_nos, rD, ref_sample[[x - 1, y - 1, z - 1]], save,
                           interactive, xlabel, ylabel, zlabel, None, savename,
                           file_extension, markersize=markersize)
            else:
                scatter_3D(temp_obj, sample_nos, rD, ref_sample, save,
                           interactive, xlabel, ylabel, zlabel, None, savename,
                           file_extension, markersize=markersize)


def show_data_domain_multi(sample_disc, Q_ref=None, Q_nums=None,
                           img_folder='figs/', ref_markers=None, ref_colors=None, showdim=None,
                           file_extension=".png", markersize=75):
    r"""
    Plots 2-D projections of the data domain D using a triangulation based on
    the first two coordinates (parameters) of the generating samples where
    :math:`Q={q_1, q_i}` for ``i=Q_nums``, with a marker for various
    :math:`Q_{ref}`.

    :param sample_disc: Object containing the samples to plot
    :type sample_disc: :class:`~bet.sample.discretization`
    :param Q_ref: reference data value
    :type Q_ref: :class:`numpy.ndarray` of shape (M, mdim)
    :param list Q_nums: dimensions of the QoI to plot
    :param string img_folder: folder to save the plots to
    :param list ref_markers: list of marker types for :math:`Q_{ref}`
    :param list ref_colors: list of colors for :math:`Q_{ref}`
    :param showdim: default 1. If int then flag to show all combinations with a
        given dimension (:math:`q_i`) or if ``all`` show all combinations.
    :type showdim: int or string
    :param string file_extension: file extension

    """
    if not isinstance(sample_disc, sample.discretization):
        raise bad_object("Improper sample object")

    # Set the default marker and colors
    if ref_markers is None:
        ref_markers = markers
    if ref_colors is None:
        ref_colors = colors

    data_obj = sample_disc._output_sample_set
    sample_obj = sample_disc._input_sample_set

    if Q_ref is None:
        Q_ref = data_obj._reference_value

    # If no specific coordinate numbers are given for the data coordinates
    # (e.g. i, where \q_i is a coordinate in the data space), then
    # set them to be the the counting numbers.
    if Q_nums is None:
        Q_nums = list(range(data_obj.get_dim()))

    # If no specific coordinate number of choice is given set to be the first
    # coordinate direction.
    if showdim is None:
        showdim = 0

    # Create a folder for these figures if it doesn't already exist
    if not os.path.isdir(img_folder):
        os.mkdir(img_folder)

    # Make sure the shape of Q_ref is correct
    if Q_ref is not None:
        Q_ref = util.fix_dimensions_data(Q_ref, data_obj.get_dim())

    # Create the triangulization to use to define the topology of the samples
    # in the data space from the first two parameters in the parameter space
    triangulation = tri.Triangulation(sample_obj.get_values()[:, 0],
                                      sample_obj.get_values()[:, 1])
    triangles = triangulation.triangles

    # Create plots of the showdim^th QoI (q_{showdim}) with all other QoI (q_i)
    if isinstance(showdim, int):
        for i in Q_nums:
            if i != showdim:
                xlabel = r'$q_{' + str(showdim + 1) + r'}$'
                ylabel = r'$q_{' + str(i + 1) + r'}$'

                filenames = [img_folder + 'domain_q' + str(showdim + 1) + '_q' +
                             str(i + 1), img_folder + 'q' + str(showdim + 1) +
                             '_q' + str(i + 1) + '_domain_Q_cs']

                data_obj_temp = sample.sample_set(2)
                data_obj_temp.set_values(
                    data_obj.get_values()[:, [showdim, i]])
                sample_disc_temp = sample.discretization(
                    sample_obj, data_obj_temp)

                if Q_ref is not None:
                    show_data_domain_2D(sample_disc_temp, Q_ref[:, [showdim, i]],
                                        ref_markers, ref_colors, xlabel=xlabel,
                                        ylabel=ylabel, triangles=triangles,
                                        save=True, interactive=False,
                                        filenames=filenames,
                                        file_extension=file_extension,
                                        markersize=markersize)
                else:
                    show_data_domain_2D(sample_disc_temp, None, ref_markers,
                                        ref_colors, xlabel=xlabel, ylabel=ylabel,
                                        triangles=triangles, save=True,
                                        interactive=False,
                                        filenames=filenames,
                                        file_extension=file_extension,
                                        markersize=markersize)
    # Create plots of all combinations of QoI in 2D
    elif showdim == 'all' or showdim == 'ALL':
        for x, y in combinations(Q_nums, 2):
            xlabel = r'$q_{' + str(x + 1) + r'}$'
            ylabel = r'$q_{' + str(y + 1) + r'}$'

            filenames = [img_folder + 'domain_q' + str(x + 1) + '_q' +
                         str(y + 1), img_folder + 'q' + str(x + 1) + '_q' +
                         str(y + 1) + '_domain_Q_cs']

            data_obj_temp = sample.sample_set(2)
            data_obj_temp.set_values(data_obj.get_values()[:, [x, y]])
            sample_disc_temp = sample.discretization(sample_obj, data_obj_temp)

            if Q_ref is not None:
                show_data_domain_2D(sample_disc_temp, Q_ref[:, [x, y]],
                                    ref_markers, ref_colors, xlabel=xlabel, ylabel=ylabel,
                                    triangles=triangles, save=True, interactive=False,
                                    filenames=filenames,
                                    file_extension=file_extension,
                                    markersize=markersize)
            else:
                show_data_domain_2D(sample_disc_temp, None, ref_markers,
                                    ref_colors, xlabel=xlabel, ylabel=ylabel,
                                    triangles=triangles, save=True, interactive=False,
                                    filenames=filenames,
                                    file_extension=file_extension,
                                    markersize=markersize)


def show_data_domain_2D(sample_disc, Q_ref=None, ref_markers=None,
                        ref_colors=None, xlabel=r'$q_1$', ylabel=r'$q_2$',
                        triangles=None, save=True, interactive=False,
                        filenames=None, file_extension=".png", markersize=75):
    r"""
    Plots 2-D a single data domain D using a triangulation based on the first
    two coordinates (parameters) of the generating samples where :math:`Q={q_1,
    q_i}` for ``i=Q_nums``, with a marker for various :math:`Q_{ref}`. Assumes
    that the first dimension of data is :math:`q_1`.

    .. note::

        Do not specify the file extension in BOTH ``filenames`` and
        ``file_extension``.


    :param sample_disc: Object containing the samples to plot
    :type sample_disc: :class:`~bet.sample.discretization`
        or :class:`~bet.sample.sample_set_base`
    :param Q_ref: reference data value
    :type Q_ref: :class:`numpy.ndarray` of shape (M, 2)
    :param list ref_markers: list of marker types for :math:`Q_{ref}`
    :param list ref_colors: list of colors for :math:`Q_{ref}`
    :param string xlabel: x-axis label
    :param string ylabel: y-axis label
    :param triangles: triangulation defined by ``samples``
    :type triangles: :class:`tri.Triuangulation.triangles`
    :param bool save: flag whether or not to save the figure
    :param bool interactive: flag whether or not to show the figure
    :param list filenames: file names for the unmarked and marked domain plots
    :param string file_extension: file extension

    """
    if not isinstance(sample_disc, sample.discretization):
        raise bad_object("Improper sample object")

    data_obj = sample_disc._output_sample_set
    sample_obj = sample_disc._input_sample_set

    if Q_ref is None:
        Q_ref = data_obj._reference_value

    # Set the default marker and colors
    if ref_markers is None:
        ref_markers = markers
    if ref_colors is None:
        ref_colors = colors
    # If no specific coordinate numbers are given for the data coordinates
    # (e.g. i, where \q_i is a coordinate in the data space), then
    # set them to be the the counting numbers.
    if triangles is None:
        triangulation = tri.Triangulation(sample_obj.get_values()[:, 0],
                                          sample_obj.get_values()[:, 1])
        triangles = triangulation.triangles
    # Set default file names
    if filenames is None:
        filenames = ['domain_q1_q2_cs', 'q1_q2_domain_Q_cs']

    # Make sure the shape of Q_ref is correct
    if Q_ref is not None:
        Q_ref = util.fix_dimensions_data(Q_ref, 2)

    # Create figure
    plt.tricontourf(data_obj.get_values()[:, 0],
                    data_obj.get_values()[:, 1],
                    triangles,
                    np.zeros((data_obj.get_values().shape[0],)),
                    colors='grey')
    plt.autoscale(tight=True)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)

    if "." not in filenames[0]:
        full_filenames0 = filenames[0] + file_extension
    else:
        full_filenames0 = filenames[0]

    if "." not in filenames[1]:
        full_filenames1 = filenames[1] + file_extension
    else:
        full_filenames1 = filenames[1]

    plt.savefig(full_filenames0, bbox_inches='tight', transparent=True,
                pad_inches=.2)
    # Add truth markers
    if Q_ref is not None:
        for i in range(Q_ref.shape[0]):
            plt.scatter(Q_ref[i, 0], Q_ref[i, 1], s=60, c=ref_colors[i],
                        marker=ref_markers[i])
    plt.tight_layout()
    if save:
        plt.savefig(full_filenames1, bbox_inches='tight', transparent=True,
                    pad_inches=.2)
    if interactive:
        plt.show()
    else:
        plt.close()


def scatter_2D_multi(sample_obj, color=None, ref_sample=None,
                     img_folder='figs/', filename="scatter2Dm", label_char=r'$\lambda',
                     showdim=None, file_extension=".png", cbar_label=None, markersize=75):
    r"""
    Creates two-dimensional projections of scatter plots of samples colored
    by ``color`` (usually an array of pointwise probability density values). A
    reference sample (``ref_sample``) can be chosen by the user. This reference
    sample will be plotted as a mauve circle twice the size of the other
    markers.

    .. note::

        Do not specify the file extension in BOTH ``filename`` and
        ``file_extension``.

    :param sample_obj: Object containing the samples to plot
    :type sample_obj: :class:`~bet.sample.sample_set_base`
    :param color: values to color the ``samples`` by
    :type color: :class:`numpy.ndarray`
    :param string filename: filename to save the figure as
    :param string label_char: character to use to label coordinate axes
    :param bool save: flag whether or not to save the figure
    :param bool interactive: flag whether or not to show the figure
    :param string img_folder: folder to save the plots to
    :param showdim: default 1. If int then flag to show all combinations with a
        given dimension (:math:`\lambda_i`) or if ``all`` show all combinations.
    :type showdim: int or string
    :param string filename: filename to save the figure as
    :param string cbar_label: color bar label

    """
    if not isinstance(sample_obj, sample.sample_set_base):
        raise bad_object("Improper sample object")
    # If no specific coordinate number of choice is given set to be the first
    # coordinate direction.
    if showdim is None:
        showdim = 0
    # Create a folder for these figures if it doesn't already exist
    if not os.path.isdir(img_folder):
        os.mkdir(img_folder)
    # Create list of all the parameter coordinates
    p_nums = np.arange(sample_obj.get_dim())

    # Create plots of the showdim^th parameter (\lambda_{showdim}) with all the
    # other parameters
    if isinstance(showdim, int):
        for i in p_nums:
            xlabel = label_char + r'_{' + str(showdim + 1) + r'}$'
            ylabel = label_char + r'_{' + str(i + 1) + r'}$'

            postfix = '_d' + str(showdim + 1) + '_d' + str(i + 1)
            myfilename = os.path.join(img_folder, filename + postfix)

            sample_obj_temp = sample.sample_set(2)
            sample_obj_temp.set_values(
                sample_obj.get_values()[:, [showdim, i]])

            if ref_sample is not None:
                scatter_2D(sample_obj_temp, sample_nos=None, color=color,
                           ref_sample=ref_sample[[showdim, i]], save=True,
                           interactive=False, xlabel=xlabel, ylabel=ylabel,
                           cbar_label=cbar_label, filename=myfilename,
                           file_extension=file_extension, markersize=markersize)
            else:
                scatter_2D(sample_obj_temp, sample_nos=None,
                           color=color, ref_sample=None, save=True,
                           interactive=False, xlabel=xlabel, ylabel=ylabel,
                           cbar_label=cbar_label, filename=myfilename,
                           file_extension=file_extension, markersize=markersize)

    # Create plots of all of the possible pairwise combinations of parameters
    elif showdim == 'all' or showdim == 'ALL':
        for x, y in combinations(p_nums, 2):
            xlabel = label_char + r'_{' + str(x + 1) + r'}$'
            ylabel = label_char + r'_{' + str(y + 1) + r'}$'

            postfix = '_d' + str(x + 1) + '_d' + str(y + 1)
            myfilename = os.path.join(img_folder, filename + postfix)

            sample_obj_temp = sample.sample_set(2)
            sample_obj_temp.set_values(sample_obj.get_values()[:, [x, y]])

            if ref_sample is not None:
                scatter_2D(sample_obj_temp, sample_nos=None, color=color,
                           ref_sample=ref_sample[[x, y]], save=True,
                           interactive=False, xlabel=xlabel, ylabel=ylabel,
                           cbar_label=cbar_label, filename=myfilename, markersize=markersize,
                           file_extension=file_extension)
            else:
                scatter_2D(sample_obj_temp, sample_nos=None, color=color,
                           ref_sample=None, save=True, interactive=False,
                           xlabel=xlabel, ylabel=ylabel,
                           cbar_label=cbar_label, markersize=markersize,
                           filename=myfilename, file_extension=file_extension)


def scatter_2D_multi_input(my_disc, color=None, ref_sample=None,
                           img_folder='figs/', filename="scatter2Dm_input",
                           label_char=r'$\lambda', showdim=None, file_extension=".png"):
    r"""
    Creates two-dimensional projections of scatter plots of samples colored
    by ``color`` (usually an array of pointwise probability density values). A
    reference sample (``ref_sample``) can be chosen by the user. This reference
    sample will be plotted as a mauve circle twice the size of the other
    markers.

    .. note::

        Do not specify the file extension in BOTH ``filename`` and
        ``file_extension``.

    :param my_disc: contains samples (`my_disc._output_sample_set``) to create
        scatter plot
    :type my_disc: :class:`~bet.sample.discretization`
    :param color: values to color the ``samples`` by
    :type color: :class:`numpy.ndarray` or string (volumes, probabilities,
        radii, normalized radii, or error id)
    :param string filename: filename to save the figure as
    :param string label_char: character to use to label coordinate axes
    :param bool save: flag whether or not to save the figure
    :param bool interactive: flag whether or not to show the figure
    :param string img_folder: folder to save the plots to
    :param showdim: default 1. If int then flag to show all combinations with a
        given dimension (:math:`\lambda_i`) or if ``all`` show all combinations.
    :type showdim: int or string
    :param string filename: filename to save the figure as

    """
    if not isinstance(my_disc, sample.discretization):
        raise bad_object("Improper sample object")

    sample_obj = my_disc.get_input_sample_set()

    if color is "volumes":
        cbar_label = r'$\mu_\Lambda(\mathcal{V}_{i,N})$'
        color = sample_obj.get_volumes()
    elif color is "probabilities":
        cbar_label = r'$\P_\Lambda(\mathcal{V}_{i,N})$'
        color = sample_obj.get_probabilities()
    elif color is "radii":
        cbar_label = r'$diam_\Lambda(\mathcal{V}_{i,N})$'
        color = sample_obj._radii
    elif color is "normalized radii":
        cbar_label = r'$diam_\Lambda_{norm}(\mathcal{V}_{i,N})$'
        color = sample_obj._normalized_radii
    elif color is "error_id":
        cbar_label = 'error id'
        color = sample_obj.get_error_id()

    scatter_2D_multi(sample_obj, color=color, ref_sample=ref_sample,
                     img_folder=img_folder, filename=filename, label_char=label_char,
                     showdim=showdim, file_extension=file_extension,
                     cbar_label=cbar_label, markersize=markersize)


def scatter_2D_multi_output(my_disc, color=None, ref_sample=None,
                            img_folder='figs/', filename="scatter2Dm_output",
                            label_char=r'$q$', showdim=None, file_extension=".png", markersize=75):
    r"""
    Creates two-dimensional projections of scatter plots of samples colored
    by ``color`` (usually an array of pointwise probability density values). A
    reference sample (``ref_sample``) can be chosen by the user. This reference
    sample will be plotted as a mauve circle twice the size of the other
    markers.

    .. note::

        Do not specify the file extension in BOTH ``filename`` and
        ``file_extension``.

    :param my_disc: contains samples (`my_disc._output_sample_set``) to create
        scatter plot
    :type my_disc: :class:`~bet.sample.discretization`
    :param color: values to color the ``samples`` by
    :type color: :class:`numpy.ndarray` or string (volumes, probabilities,
        radii, normalized radii, or error id)
    :param string filename: filename to save the figure as
    :param string label_char: character to use to label coordinate axes
    :param bool save: flag whether or not to save the figure
    :param bool interactive: flag whether or not to show the figure
    :param string img_folder: folder to save the plots to
    :param showdim: default 1. If int then flag to show all combinations with a
        given dimension (:math:`q_i`) or if ``all`` show all combinations.
    :type showdim: int or string
    :param string filename: filename to save the figure as

    """
    if not isinstance(my_disc, sample.discretization):
        raise bad_object("Improper sample object")

    sample_obj = my_disc.get_output_sample_set()

    if color is "volumes":
        cbar_label = r'$\mu_\mathcal{D}(\mathcal{I}_{j,M})$'
        color = sample_obj.get_volumes()
    elif color is "probabilities":
        cbar_label = r'$\P_\mathcal{D}(\mathcal{I}_{j,M})$'
        color = sample_obj.get_probabilities()
    elif color is "radii":
        cbar_label = r'$diam_\mathcal{D}(\mathcal{I}_{j,M})$'
        color = sample_obj._radii
    elif color is "normalized radii":
        cbar_label = r'$diam_\mathcal{D}_{norm}(\mathcal{I}_{j,M})$'
        color = sample_obj._normalized_radii
    elif color is "error_id":
        cbar_label = 'error id'
        color = sample_obj.get_error_id()

    scatter_2D_multi(sample_obj, color=color, ref_sample=ref_sample,
                     img_folder=img_folder, filename=filename, label_char=label_char,
                     showdim=showdim, file_extension=file_extension,
                     cbar_label=cbar_label, markersize=markersize)



# Copyright (C) 2014-2020 The BET Development Team

"""
This example generates uniform random samples in the unit hypercube and
corresponding QoIs (data) generated by a linear map Q.  We then calculate the
gradients using an RBF scheme and use the gradient information to choose the
optimal set of 2 (3, 4, ... input_dim) QoIs to use in the inverse problem.

Every real world problem requires special attention regarding how we choose
*optimal QoIs*.  This set of examples (examples/sensitivity/linear) covers
some of the more common scenarios using easy to understand linear maps.

In this *measure_binsize_large* example we choose *optimal QoIs* to be the set of
QoIs of size input_dim that produces the smallest support of the inverse
solution, assuming we define the uncertainty in our data to be fixed, i.e.,
independent of the range of data maesured for each QoI (bin_size).
"""

import numpy as np
import bet.sensitivity.gradients as grad
import bet.sensitivity.chooseQoIs as cqoi
import bet.calculateP.simpleFunP as simpleFunP
import bet.calculateP.calculateP as calculateP
import bet.postProcess.postTools as postTools
import bet.Comm as comm
import bet.sample as sample

# Set up the info for the spaces
input_dim = 10
output_dim = 100
num_samples = 1E5
num_centers = 10

# Let the map Q be a random matrix of size (output_dim, input_dim)
np.random.seed(0)
Q = np.random.random([output_dim, input_dim])

# Initialize some sample objects we will need
input_samples = sample.sample_set(input_dim)
output_samples = sample.sample_set(output_dim)

# Choose random samples in parameter space to solve the model
input_samples.set_values(np.random.uniform(
    0, 1, [np.int(num_samples), input_dim]))

# Make the MC assumption and compute the volumes of each Voronoi cell
input_samples.estimate_volume_mc()


# Compute the output values with the map Q
output_samples.set_values(Q.dot(input_samples.get_values().transpose()).
                          transpose())

# Calculate the gradient vectors at some subset of the samples.  Here the
# *normalize* argument is set to *False* because we are using bin_size to
# determine the uncertainty in our data.
cluster_discretization = sample.discretization(input_samples, output_samples)
# We will approximate the jacobian at each of the centers
center_discretization = grad.calculate_gradients_rbf(cluster_discretization,
                                                     num_centers, normalize=False)

# With these gradient vectors, we are now ready to choose an optimal set of
# QoIs to use in the inverse problem, based on minimizing the support of the
# inverse solution (volume).  The most robust method for this is
# :meth:~bet.sensitivity.chooseQoIs.chooseOptQoIs_large which returns the
# best set of 2, 3, 4 ... until input_dim.  This method returns a list of
# matrices.  Each matrix has 10 rows, the first column representing the
# expected inverse volume ratio, and the rest of the columns the corresponding
# QoI indices.
input_samples_center = center_discretization.get_input_sample_set()
best_sets = cqoi.chooseOptQoIs_large(input_samples_center, max_qois_return=5,
                                     num_optsets_return=2, inner_prod_tol=0.9, measskew_tol=1E2, measure=True)

'''
We see here the expected volume ratios are small.  This number represents the
expected volume of the inverse image of a unit hypercube in the data space.
With the bin_size definition of the uncertainty in the data, here we expect to
see inverse solutions that have a smaller support (expected volume ratio < 1)
than the original volume of the hypercube in the data space.

This interpretation of the expected volume ratios is only valid for inverting
from a data space that has the same dimensions as the parameter space.  When
inverting into a higher dimensional space, this expected volume ratio is the
expected volume of the cross section of the inverse solution.
'''
###############################################################################

# At this point we have determined the optimal set of QoIs to use in the inverse
# problem.  Now we compare the support of the inverse solution using
# different sets of these QoIs.  We set Q_ref to correspond to the center of
# the parameter space.  We choose the set of QoIs to consider.

QoI_indices = [0, 7]  # choose up to input_dim
QoI_indices = [0, 1]
QoI_indices = [0, 7, 34, 39, 90]
#QoI_indices = [0, 1, 2, 3, 4]

# Choose some QoI indices to solve the inverse problem with
output_samples._dim = len(QoI_indices)
output_samples.set_values(output_samples.get_values()[:, QoI_indices])

# Define the reference point in the output space to correspond to the center of
# the input space.
param_ref = 0.5 * np.ones(input_dim)
Q_ref = Q[QoI_indices, :].dot(param_ref)

# bin_ratio defines the uncertainty in our data
bin_ratio = 0.25

# Create discretization object
my_discretization = sample.discretization(input_sample_set=input_samples,
                                          output_sample_set=output_samples)


# Find the simple function approximation
simpleFunP.regular_partition_uniform_distribution_rectangle_scaled(
    data_set=my_discretization, Q_ref=Q_ref, rect_scale=bin_ratio,
    cells_per_dimension=1)

# Calculate probabilities making the Monte Carlo assumption
calculateP.prob(my_discretization)

percentile = 1.0
# Sort samples by highest probability density and find how many samples lie in
# the support of the inverse solution.  With the Monte Carlo assumption, this
# also tells us the approximate volume of this support.
(num_samples, _, indices_in_inverse) =\
    postTools.sample_highest_prob(top_percentile=percentile,
                                  sample_set=input_samples, sort=True)

# Print the approximate proportion of the measure of the parameter space defined
# by the support of the inverse density
if comm.rank == 0:
    print('The approximate proportion of the measure of the parameter space defined')
    print('by the support of the inverse density associated with the choice of QoI map is')
    print(np.sum(input_samples.get_volumes()[indices_in_inverse]),
          ' with ', num_samples, ' samples.')



# Copyright (C) 2014-2020 The BET Development Team

"""
This module contains functions for sampling. We assume we are given access to a
model, a parameter space, and a data space. The model is a map from the
parameter space to the data space. We desire to build up a set of samples to
solve an inverse problem thus giving information about the inverse mapping.
Each sample consists for a parameter coordinate, data coordinate pairing. We
assume the measure on both spaces is Lebesgue.
"""

import collections.abc
import os
import warnings
import logging
import glob
import numpy as np
import scipy.io as sio
import scipy.stats as stats
from pyDOE import lhs
from bet.Comm import comm
import bet.sample as sample
import bet.sample
import bet.util as util


class bad_object(Exception):
    """
    Exception for when the wrong type of object is used.
    """


def sample_from_updated(input_set, num_samples, globalize=True):
    """
    Create a new sample set from resampling from the updated probability measure of another sample set.

    :param input_set: Sample set or discretization containing updated probability measure from which to sample.
    :type input_set: :class:`~bet.sample.sample_set` or :class:`~bet.sample.discretization`
    :param num_samples: Number of new samples to create.
    :type num_samples: int
    :param globalize: Whether or not to globalize objects.
    :type bool
    :return: Sample set containing new samples
    :rtype: :class:`~bet.sample.sample_set`

    """
    if isinstance(input_set, bet.sample.discretization):
        input_set = input_set.get_input_sample_set()
    elif not isinstance(input_set, bet.sample.sample_set):
        raise bad_object("input_set is of the wrong type.")

    new_set = sample.sample_set(dim=input_set.get_dim())
    if input_set.get_prob_type() == 'rv':
        return random_sample_set(input_set.get_prob_parameters(), new_set, num_samples, globalize)
    elif input_set.get_prob_type() == 'kde':
        param_marginals, cluster_weights = input_set.get_prob_parameters()
        v_outer = []
        for i, w in enumerate(cluster_weights):
            v_inner = []
            num_samples_clust = round(w*num_samples)
            num_samples_local = int((num_samples_clust / comm.size) +
                                    (comm.rank < num_samples_clust % comm.size))
            for j in range(input_set.get_dim()):
                v_inner.append(param_marginals[j][i].resample(num_samples_local))
            v_outer.append(np.vstack(v_inner))
        vals_local = np.hstack(v_outer)
        new_set.set_values_local(vals_local)
        new_set.set_prob_type_init('kde')
        new_set.set_prob_parameters_init((param_marginals, cluster_weights))
        if globalize:
            new_set.local_to_global()
        return new_set
    elif input_set.get_prob_type() == 'gmm':
        means, covariances, cluster_weights = input_set.get_prob_parameters()
        v_outer = []
        for i, w in enumerate(cluster_weights):
            num_samples_clust = round(w * num_samples)
            num_samples_local = int((num_samples_clust / comm.size) +
                                    (comm.rank < num_samples_clust % comm.size))
            v_outer.append(stats.multivariate_normal.rvs(mean=means[i], cov=covariances[i], size=num_samples_local))
        vals_local = np.vstack(v_outer)
        new_set.set_values_local(vals_local)
        new_set.set_prob_type_init('gmm')
        new_set.set_prob_parameters_init((means, covariances, cluster_weights))
        if globalize:
            new_set.local_to_global()
        return new_set
    else:
        raise bad_object("The updated probability measure is undefined or not allowed for this method.")


def random_sample_set(rv, input_obj, num_samples, globalize=True):
    """
    Create a sample set by sampling random variates from continuous distributions
    from :class:`scipy.stats.rv_continuous`. See https://docs.scipy.org/doc/scipy/reference/stats.html.

    `rv` can take multiple types of formats depending on type of distribution.

    A string is used for the same distribution with default parameters in each dimension.
    ex. rv = 'uniform' or rv = 'beta'

    A list or tuple of length 2 is used for the same distribution with user-defined parameters in each dimension as a
    dictionary.
    ex. rv = ['uniform', {'loc':-2, 'scale':5}] or rv = ['beta', {'a': 2, 'b':5, 'loc':-2, 'scale':5}]

    A list of length dim which entries of lists or tuples of length 2 is used for different distributions with
    user-defined parameters in each dimension as a
    dictionary.
    ex. rv = [['uniform', {'loc':-2, 'scale':5}],
              ['beta', {'a': 2, 'b':5, 'loc':-2, 'scale':5}]]

    :param rv: Type and parameters for continuous random variables.
    :type rv: str, list, or tuple
    :param input_obj: :class:`~bet.sample.sample_set` object containing the dimension to sample from, or the dimension.
    :type input_obj: :class:`~bet.sample.sample_set` or int or :class:`numpy.ndarray`
    :param num_samples: Number of samples
    :type num_samples: int
    :param globalize: Whether or not to globalize vectors.
    :type globalize: bool

    """
    # for backward compatibility
    if rv == "r" or rv == "random":
        rv = "uniform"
    elif rv == 'lhs':
        return lhs_sample_set(input_obj, num_samples, criterion='center', globalize=globalize)
    # check to see what the input object is
    if isinstance(input_obj, sample.sample_set):
        input_sample_set = input_obj
    elif isinstance(input_obj, int):
        input_sample_set = sample.sample_set(input_obj)
    elif isinstance(input_obj, np.ndarray):
        input_sample_set = sample.sample_set(input_obj.shape[0])
        input_sample_set.set_domain(input_obj)
    else:
        raise sample.wrong_input("input_obj is of wrong type.")

    dim = input_sample_set.get_dim()

    if type(rv) is str:
        if input_sample_set.get_domain() is None:
            rv = [[rv, {}]] * dim
        else:
            domain = input_sample_set.get_domain()
            rv_type = rv
            rv = []
            for i in range(dim):
                rv.append([rv_type, {'loc': domain[i, 0], 'scale': domain[i, 1]-domain[i, 0]}])
    elif type(rv) in (list, tuple):
        if len(rv) == 2 and type(rv[0]) is str and type(rv[1]) is dict:
            rv = [rv] * dim
        elif len(rv) != dim:
            raise sample.dim_not_matching("rv has fewer entries than the dimension.")
    else:
        raise sample.wrong_input("rv must be a string, list, or tuple.")

    # define local number of samples
    num_samples_local = int((num_samples / comm.size) +
                            (comm.rank < num_samples % comm.size))

    input_values_local = np.empty((num_samples_local, dim))
    domain = np.empty((dim, 2))

    for i in range(dim):
        rv_continuous = getattr(stats, rv[i][0])
        args = rv[i][1]
        input_values_local[:, i] = rv_continuous.rvs(size=num_samples_local, **args)
        domain[i, :] = rv_continuous.interval(1, **args)
    input_sample_set.set_values_local(input_values_local)
    input_sample_set.set_domain(domain)
    input_sample_set.set_prob_type_init("rv")
    input_sample_set.set_prob_parameters_init(rv)
    input_sample_set.check_num_local()
    input_sample_set.check_num()

    comm.barrier()

    if globalize:
        input_sample_set.local_to_global()
    else:
        input_sample_set._values = None
    return input_sample_set


def lhs_sample_set(input_obj, num_samples, criterion, globalize=True):
    """
    Sampling algorithm for generating samples from a Latin hypercube
    in the domain present with ``input_obj`` (a default unit hypercube
    is used if no domain has been specified)

    :param input_obj: :class:`~bet.sample.sample_set` object containing
        the dimension or domain to sample from, the domain to sample from, or
        the dimension
    :type input_obj: :class:`~bet.sample.sample_set` or :class:`numpy.ndarray`
        of shape (dim, 2) or ``int``
    :param num_samples: number of samples
    :type num_samples: int
    :param criterion: latin hypercube criterion see
            `PyDOE <http://pythonhosted.org/pyDOE/randomized.html>`
    :type criterion: str
    :param globalize: Whether or not to globalize local variables.
    :type globalize: bool
    :rtype: :class:`~bet.sample.sample_set`
    :returns: :class:`~bet.sample.sample_set`

    """
    # check to see what the input object is
    if isinstance(input_obj, sample.sample_set):
        input_sample_set = input_obj
    elif isinstance(input_obj, int):
        input_sample_set = sample.sample_set(input_obj)
    elif isinstance(input_obj, np.ndarray):
        input_sample_set = sample.sample_set(input_obj.shape[0])
        input_sample_set.set_domain(input_obj)

    dim = input_sample_set.get_dim()
    if input_sample_set.get_domain() is None:
        # create the domain
        input_domain = np.array([[0., 1.]] * dim)
        input_sample_set.set_domain(input_domain)
        logging.warning("Setting domain to hypercube.")

    # update the bounds based on the number of samples
    input_sample_set.update_bounds(num_samples)
    input_values = np.copy(input_sample_set._width)
    input_values = input_values * lhs(dim, num_samples, criterion)
    input_values = input_values + input_sample_set._left
    input_sample_set.set_values_local(np.array_split(input_values, comm.size)[comm.rank])

    comm.barrier()
    if globalize:
        input_sample_set.local_to_global()
    else:
        input_sample_set._values = None
    input_sample_set.set_prob_type_init("lhs")
    input_sample_set.set_prob_parameters_init(criterion)

    return input_sample_set


def regular_sample_set(input_obj, num_samples_per_dim=1):
    """
    Sampling algorithm for generating a regular grid of samples taken
    on the domain present with ``input_obj`` (a default unit hypercube
    is used if no domain has been specified)

    :param input_obj: :class:`~bet.sample.sample_set` object containing
        the dimension or domain to sample from, the domain to sample from, or
        the dimension
    :type input_obj: :class:`~bet.sample.sample_set` or :class:`numpy.ndarray`
        of shape (dim, 2) or ``int``
    :param num_samples_per_dim: number of samples per dimension
    :type num_samples_per_dim: :class:`~numpy.ndarray` of dimension
        ``(input_sample_set._dim,)``

    :rtype: :class:`~bet.sample.sample_set`
    :returns: :class:`~bet.sample.sample_set` object which contains
        input ``num_samples``

    """
    # check to see what the input object is
    if isinstance(input_obj, sample.sample_set):
        input_sample_set = input_obj.copy()
    elif isinstance(input_obj, int):
        input_sample_set = sample.sample_set(input_obj)
    elif isinstance(input_obj, np.ndarray):
        input_sample_set = sample.sample_set(input_obj.shape[0])
        input_sample_set.set_domain(input_obj)
    else:
        raise bad_object("Improper sample object")

    # Create N samples
    dim = input_sample_set.get_dim()

    if not isinstance(num_samples_per_dim, collections.abc.Iterable):
        num_samples_per_dim = num_samples_per_dim * np.ones((dim,))
    if np.any(np.less_equal(num_samples_per_dim, 0)):
        warnings.warn('Warning: num_samples_per_dim must be greater than 0')

    num_samples = int(np.product(num_samples_per_dim))

    if input_sample_set.get_domain() is None:
        # create the domain
        input_domain = np.array([[0., 1.]] * dim)
        input_sample_set.set_domain(input_domain)
    else:
        input_domain = input_sample_set.get_domain()
    # update the bounds based on the number of samples
    input_values = np.zeros((num_samples, dim))

    vec_samples_dimension = np.empty((dim), dtype=object)
    for i in range(dim):
        bin_width = (input_domain[i, 1] - input_domain[i, 0]) / \
            np.float(num_samples_per_dim[i])
        vec_samples_dimension[i] = list(np.linspace(
            input_domain[i, 0] - 0.5 * bin_width,
            input_domain[i, 1] + 0.5 * bin_width,
            int(num_samples_per_dim[i]) + 2))[1:int(num_samples_per_dim[i] + 1)]

    arrays_samples_dimension = np.meshgrid(
        *[vec_samples_dimension[i] for i in np.arange(0, dim)],
        indexing='ij')

    for i in range(dim):
        input_values[:, i:i + 1] = np.vstack(arrays_samples_dimension[i]
                                             .flat[:])

    input_sample_set.set_values(input_values)
    input_sample_set.global_to_local()
    input_sample_set.set_prob_type_init("grid")
    input_sample_set.set_prob_parameters_init(num_samples_per_dim)

    return input_sample_set


class sampler(object):
    """
    This class provides methods for sampling of parameter space to
    provide samples to be used by algorithms to solve inverse problems.

    """
    def __init__(self, lb_model,
                 error_estimates=False, jacobians=False):
        """
        Initialization

        :param lb_model: Interface to physics-based model takes an input of
            shape (N, ndim) and returns an output of shape (N, mdim)
        :type lb_model: callable function
        :param bool error_estimates: Whether or not the model returns error estimates
        :param bool jacobians: Whether or not the model returns Jacobians
        """
        self.lb_model = lb_model
        self.error_estimates = error_estimates
        self.jacobians = jacobians
        self.input_sample_set = None
        self.discretization = None

    def local_to_global(self):
        """
        Globalize local variables.
        """
        if self.input_sample_set is not None:
            self.input_sample_set.local_to_global()
        if self.discretization is not None:
            self.discretization.local_to_global()

    def random_sample_set(self, rv, input_obj, num_samples, globalize=True):
        """
        Create a sample set by sampling random variates from continuous distributions
        from :class:`scipy.stats.rv_continuous`. See https://docs.scipy.org/doc/scipy/reference/stats.html.

        `rv` can take multiple types of formats depending on type of distribution.

        A string is used for the same distribution with default parameters in each dimension.
        ex. rv = 'uniform' or rv = 'beta'

        A list or tuple of length 2 is used for the same distribution with user-defined parameters in each dimension as a
        dictionary.
        ex. rv = ['uniform', {'loc':-2, 'scale':5}] or rv = ['beta', {'a': 2, 'b':5, 'loc':-2, 'scale':5}]

        A list of length dim which entries of lists or tuples of length 2 is used for different distributions with
        user-defined parameters in each dimension as a
        dictionary.
        ex. rv = [['uniform', {'loc':-2, 'scale':5}],
                  ['beta', {'a': 2, 'b':5, 'loc':-2, 'scale':5}]]

        :param rv: Type and parameters for continuous random variables.
        :type rv: str, list, or tuple
        :param input_obj: :class:`~bet.sample.sample_set` object containing the dimension to sample from, or the dimension.
        :type input_obj: :class:`~bet.sample.sample_set` or int
        :param num_samples: Number of samples
        :type num_samples: int
        :param globalize: Whether or not to globalize vectors.
        :type globalize: bool
        :return:
        """
        self.input_sample_set = random_sample_set(rv, input_obj, num_samples, globalize=globalize)
        return self.input_sample_set

    def regular_sample_set(self, input_obj, num_samples_per_dim=1):
        """
        Sampling algorithm for generating a regular grid of samples taken
        on the domain present with ``input_obj`` (a default unit hypercube
        is used if no domain has been specified)

        :param input_obj: :class:`~bet.sample.sample_set` object containing
            the dimension or domain to sample from, the domain to sample from, or
            the dimension
        :type input_obj: :class:`~bet.sample.sample_set` or :class:`numpy.ndarray`
            of shape (dim, 2) or ``int``
        :param num_samples_per_dim: number of samples per dimension
        :type num_samples_per_dim: :class:`~numpy.ndarray` of dimension
            ``(input_sample_set._dim,)``

        :rtype: :class:`~bet.sample.sample_set`
        :returns: :class:`~bet.sample.sample_set` object which contains
            input ``num_samples``

        """
        self.input_sample_set = regular_sample_set(input_obj, num_samples_per_dim)
        return self.input_sample_set

    def lhs_sample_set(self, input_obj, num_samples, criterion, globalize=True):
        """
        Sampling algorithm for generating samples from a Latin hypercube
        in the domain present with ``input_obj`` (a default unit hypercube
        is used if no domain has been specified)

        :param input_obj: :class:`~bet.sample.sample_set` object containing
            the dimension or domain to sample from, the domain to sample from, or
            the dimension
        :type input_obj: :class:`~bet.sample.sample_set` or :class:`numpy.ndarray`
            of shape (dim, 2) or ``int``
        :param num_samples: number of samples
        :type num_samples: int
        :param criterion: latin hypercube criterion see
                `PyDOE <http://pythonhosted.org/pyDOE/randomized.html>`
        :type criterion: str
        :param globalize: Whether or not to globalize local variables.
        :type globalize: bool
        :rtype: :class:`~bet.sample.sample_set`
        :returns: :class:`~bet.sample.sample_set`

        """
        self.input_sample_set = lhs_sample_set(input_obj, num_samples, criterion, globalize)
        return self.input_sample_set

    def compute_QoI_and_create_discretization(self, input_sample_set=None,
                                              savefile=None, globalize=True):
        """
        Dummy function for `compute_qoi_and_create_discretization`.
        """
        logging.warning("This will be removed in a later version. Use compute_qoi_and_create_discretization instead.")
        return self.compute_qoi_and_create_discretization(input_sample_set, savefile, globalize)

    def compute_qoi_and_create_discretization(self, input_sample_set=None,
                                              savefile=None, globalize=True):
        """
        Samples the model at ``input_sample_set`` and saves the results.

        Note: There are many ways to generate samples on a regular grid in
        Numpy and other Python packages. Instead of reimplementing them here we
        provide sampler that utilizes user specified samples.

        :param input_sample_set: samples to evaluate the model at
        :type input_sample_set: :class:`~bet.sample.sample_set` with
            num_samples
        :param string savefile: filename to save samples and data
        :param bool globalize: Makes local variables global.

        :rtype: :class:`~bet.sample.discretization`
        :returns: :class:`~bet.sample.discretization` object which contains
            input and output of length ``num_samples``

        """

        if input_sample_set is not None:
            self.input_sample_set = input_sample_set

        # Solve the model at the samples
        if self.input_sample_set._values_local is None:
            self.input_sample_set.global_to_local()

        local_output = self.lb_model(
            self.input_sample_set.get_values_local())

        if isinstance(local_output, np.ndarray):
            local_output_values = local_output
        elif isinstance(local_output, tuple):
            if len(local_output) == 1:
                local_output_values = local_output[0]
            elif len(local_output) == 2 and self.error_estimates:
                (local_output_values, local_output_ee) = local_output
            elif len(local_output) == 2 and self.jacobians:
                (local_output_values, local_output_jac) = local_output
            elif len(local_output) == 3:
                (local_output_values, local_output_ee, local_output_jac) = \
                    local_output
        else:
            raise bad_object("lb_model is not returning the proper type")

        # figure out the dimension of the output
        if len(local_output_values.shape) <= 1:
            output_dim = 1
        else:
            output_dim = local_output_values.shape[1]

        output_sample_set = sample.sample_set(output_dim)
        output_sample_set.set_values_local(local_output_values)
        lam_ref = self.input_sample_set.get_reference_value()

        if lam_ref is not None:
            try:
                if not isinstance(lam_ref, collections.abc.Iterable):
                    lam_ref = np.array([lam_ref])
                Q_ref = self.lb_model(lam_ref)
                output_sample_set.set_reference_value(Q_ref)
            except ValueError:
                try:
                    msg = "Model not mapping reference value as expected."
                    msg += "Attempting reshape..."
                    logging.log(20, msg)
                    q_ref = self.lb_model(lam_ref.reshape(1, -1))
                    output_sample_set.set_reference_value(q_ref)
                except ValueError:
                    logging.log(20, 'Unable to map reference value.')

        if self.error_estimates:
            output_sample_set.set_error_estimates_local(local_output_ee)

        if self.jacobians:
            self.input_sample_set.set_jacobians_local(local_output_jac)

        if globalize:
            self.input_sample_set.local_to_global()
            output_sample_set.local_to_global()
        else:
            self.input_sample_set._values = None

        comm.barrier()

        self.discretization = sample.discretization(self.input_sample_set,
                                                    output_sample_set)
        comm.barrier()

        if savefile is not None:
            self.discretization.save(filename=savefile, globalize=globalize)

        comm.barrier()

        return self.discretization

    def copy(self):
        """
        Returns a copy of the sampler object.
        """
        import copy
        return copy.deepcopy(self)

    def create_random_discretization(self, rv, input_obj,
                                     savefile=None, num_samples=None,
                                     globalize=True):
        """
        Create a sample set by sampling random variates from continuous distributions
        from :class:`scipy.stats.rv_continuous`. See https://docs.scipy.org/doc/scipy/reference/stats.html,
        and evaluate the model to calculate quantities of interest and make a discretization.

        `rv` can take multiple types of formats depending on type of distribution.

        A string is used for the same distribution with default parameters in each dimension.
        ex. rv = 'uniform' or rv = 'beta'

        A list or tuple of length 2 is used for the same distribution with user-defined parameters in each dimension as a
        dictionary.
        ex. rv = ['uniform', {'loc':-2, 'scale':5}] or rv = ['beta', {'a': 2, 'b':5, 'loc':-2, 'scale':5}]

        A list of length dim which entries of lists or tuples of length 2 is used for different distributions with
        user-defined parameters in each dimension as a
        dictionary.
        ex. rv = [['uniform', {'loc':-2, 'scale':5}],
                  ['beta', {'a': 2, 'b':5, 'loc':-2, 'scale':5}]]

        :param rv: Type and parameters for continuous random variables.
        :type rv: str, list, or tuple
        :param input_obj: :class:`~bet.sample.sample_set` object containing the dimension to sample from, or the dimension.
        :type input_obj: :class:`~bet.sample.sample_set` or int
        :param string savefile: filename to save discretization
        :param num_samples: Number of samples
        :type num_samples: int
        :param globalize: Whether or not to globalize vectors.
        :type globalize: bool

        :rtype: :class:`~bet.sample.discretization`
        :returns: :class:`~bet.sample.discretization` object which contains
            input and output sample sets with ``num_samples`` total samples

        """
        # Create N samples
        if num_samples is None:
            num_samples = self.num_samples

        input_sample_set = self.random_sample_set(rv, input_obj,
                                                  num_samples, globalize)

        return self.compute_qoi_and_create_discretization(input_sample_set,
                                                          savefile, globalize)



# Copyright (C) 2014-2020 The BET Development Team

"""
This module contains tests for :module:`bet.sensitivity.chooseQoIs`.
Most of these tests should make sure certain values are within a tolerance
rather than exact due to machine precision.
"""
import unittest
import bet.sensitivity.gradients as grad
import bet.sensitivity.chooseQoIs as cQoIs
import numpy as np
import numpy.testing as nptest
from itertools import combinations
import sys
import bet.sample as sample


class ChooseQoIsMethods:
    """
    Test :module:`bet.sensitivity.chooseQoIs`.
    """

    def test_calculate_avg_measure(self):
        """
        Test :meth:`bet.sensitivity.chooseQoIs.calculate_avg_measure`.
        """
        self.qoi_set = np.arange(0, self.input_dim)
        (self.measure, self.singvals) = cQoIs.calculate_avg_measure(
            self.input_set_centers, self.qoi_set)

        # Check that measure and singvals are the right size
        self.assertEqual(isinstance(self.measure, float), True)
        self.assertEqual(self.singvals.shape, (self.num_centers,
                                               self.input_dim))

        # Test the method returns an error when more qois are given than
        # parameters
        self.input_set_centers.set_jacobians(
            np.random.uniform(-1, 1, [10, 4, 3]))
        with self.assertRaises(ValueError):
            cQoIs.calculate_avg_measure(self.input_set_centers)

    def test_calculate_avg_skewness(self):
        """
        Test :meth:`bet.sensitivity.chooseQoIs.calculate_avg_skewness`.
        """
        self.qoi_set = np.arange(0, self.input_dim)
        (self.skewness, self.skewnessgi) = cQoIs.calculate_avg_skewness(
            self.input_set_centers, self.qoi_set)

        # Check that skewness and skewnessgi are the right size
        self.assertEqual(isinstance(self.skewness, float), True)
        self.assertEqual(self.skewnessgi.shape, (self.num_centers,
                                                 self.input_dim))

        # Test the method returns an error when more qois are given than
        # parameters
        self.input_set_centers.set_jacobians(
            np.random.uniform(-1, 1, [10, 4, 3]))
        with self.assertRaises(ValueError):
            cQoIs.calculate_avg_measure(self.input_set_centers)

    def test_calculate_avg_condnum(self):
        """
        Test :meth:`bet.sensitivity.chooseQoIs.calculate_avg_condnum`.
        """
        self.qoi_set = np.arange(0, self.input_dim)
        print(self.input_set_centers.get_jacobians())
        print(self.center_disc._input_sample_set.get_jacobians())
        (self.condnum, self.singvals) = cQoIs.calculate_avg_condnum(
            self.input_set_centers, self.qoi_set)

        # Check that condnum and singvals are the right size
        self.assertEqual(isinstance(self.condnum, float), True)
        self.assertEqual(self.singvals.shape, (self.num_centers,
                                               self.input_dim))

        # Test the method returns an error when more qois are given than
        # parameters
        self.input_set_centers.set_jacobians(
            np.random.uniform(-1, 1, [10, 4, 3]))
        with self.assertRaises(ValueError):
            cQoIs.calculate_avg_measure(self.input_set_centers)

    def test_chooseOptQoIs(self):
        """
        Test :meth:`bet.sensitivity.chooseQoIs.chooseOptQoIs`.
        """
        self.qoiIndices = np.arange(0, self.output_dim)
        self.condnum_indices_mat = cQoIs.chooseOptQoIs(self.input_set_centers,
                                                       self.qoiIndices, self.output_dim_return, self.num_optsets_return)
        self.condnum_indices_mat_vol = cQoIs.chooseOptQoIs(self.input_set_centers,
                                                           self.qoiIndices, self.output_dim_return, self.num_optsets_return,
                                                           measure=True)

        # Test the method returns the correct size array
        self.assertEqual(self.condnum_indices_mat.shape,
                         (self.num_optsets_return, self.output_dim_return + 1))

        self.assertEqual(self.condnum_indices_mat_vol.shape,
                         (self.num_optsets_return, self.output_dim_return + 1))

        # Check that the 'global condition number' is greater than or equal to
        # 1
        nptest.assert_array_less(1.0, self.condnum_indices_mat[:, 0])

        # For measure, check that it is greater than or equal to 0
        nptest.assert_array_less(0.0, self.condnum_indices_mat_vol[:, 0])

        # Test the method returns the known best set of QoIs  (chosen to be
        # last input_dim indices)
        nptest.assert_array_less(self.output_dim - self.input_dim - 1,
                                 self.condnum_indices_mat[0, 1:])

        nptest.assert_array_less(self.output_dim - self.input_dim - 1,
                                 self.condnum_indices_mat_vol[0, 1:])

        # Test that none of the best chosen QoIs are the same
        self.assertEqual(len(np.unique(self.condnum_indices_mat[0, 1:])),
                         len(self.condnum_indices_mat[0, 1:]))

        self.assertEqual(len(np.unique(self.condnum_indices_mat[0, 1:])),
                         len(self.condnum_indices_mat_vol[0, 1:]))

        ##########
        # Test the method for a set of QoIs rather than all possible.  Choose
        # this set so that the optimal choice is not removed.
        self.qoiIndices = np.concatenate([np.arange(1, 3, 2),
                                          np.arange(4, self.output_dim)])
        self.condnum_indices_mat = cQoIs.chooseOptQoIs(self.input_set_centers,
                                                       self.qoiIndices, self.output_dim_return, self.num_optsets_return)

        self.condnum_indices_mat_vol = cQoIs.chooseOptQoIs(self.input_set_centers,
                                                           self.qoiIndices, self.output_dim_return, self.num_optsets_return,
                                                           measure=True)

        # Test the method returns the correct number of qois
        self.assertEqual(self.condnum_indices_mat.shape,
                         (self.num_optsets_return, self.output_dim_return + 1))

        self.assertEqual(self.condnum_indices_mat_vol.shape,
                         (self.num_optsets_return, self.output_dim_return + 1))

        # Check that the 'global condidtion number' is greater than or equal
        # to 1
        nptest.assert_array_less(1.0, self.condnum_indices_mat[:, 0])

        nptest.assert_array_less(0.0, self.condnum_indices_mat_vol[:, 0])

        # Test the method returns the known best set of QoIs  (chosen to be
        # last input_dim indices)
        nptest.assert_array_less(self.output_dim - self.input_dim - 1,
                                 self.condnum_indices_mat[0, 1:])

        nptest.assert_array_less(self.output_dim - self.input_dim - 1,
                                 self.condnum_indices_mat_vol[0, 1:])

        # Test that none of the best chosen QoIs are the same
        self.assertEqual(len(np.unique(self.condnum_indices_mat[0, 1:])),
                         len(self.condnum_indices_mat[0, 1:]))

        self.assertEqual(len(np.unique(self.condnum_indices_mat[0, 1:])),
                         len(self.condnum_indices_mat_vol[0, 1:]))

    def test_chooseOptQoIs_verbose(self):
        """
        Test :meth:`bet.sensitivity.chooseQoIs.chooseOptQoIs_verbose`.
        """
        self.qoiIndices = np.arange(0, self.output_dim)
        [self.condnum_indices_mat, self.optsingvals] = \
            cQoIs.chooseOptQoIs_verbose(self.input_set_centers, self.qoiIndices,
                                        self.output_dim_return, self.num_optsets_return)

        # Test that optsingvals is the right shape
        self.assertEqual(self.optsingvals.shape, ((self.num_centers,
                                                   self.output_dim_return, self.num_optsets_return)))

    def test_find_unique_vecs(self):
        """
        Test :meth:`bet.sensitivity.chooseQoIs.find_unique_vecs`.
        """
        self.qoiIndices = np.arange(0, self.output_dim)
        unique_indices = cQoIs.find_unique_vecs(self.input_set_centers,
                                                self.inner_prod_tol, self.qoiIndices)

        # Test that pairwise inner products are <= inner_prod_tol
        pairs = np.array(list(combinations(list(unique_indices), 2)))
        for pair in range(pairs.shape[0]):
            curr_set = pairs[pair]
            curr_inner_prod = np.sum(self.input_set_centers._jacobians[:,
                                                                       curr_set[0], :] * self.input_set_centers._jacobians[:,
                                                                                                                           curr_set[1], :]) / self.input_set_centers._jacobians.shape[0]
            nptest.assert_array_less(curr_inner_prod, self.inner_prod_tol)

    def test_chooseOptQoIs_large(self):
        """
        Test :meth:`bet.sensitivity.chooseQoIs.chooseOptQoIs_large`.
        """
        self.qoiIndices = np.arange(0, self.output_dim)
        best_sets = cQoIs.chooseOptQoIs_large(self.input_set_centers,
                                              qoiIndices=self.qoiIndices, inner_prod_tol=self.inner_prod_tol,
                                              measskew_tol=self.measskew_tol)

        if self.measskew_tol == np.inf:
            self.measskew_tol = sys.float_info[0]
        # Test that the best_sets have condition number less than the tolerance
        for Ldim in range(self.input_dim - 1):
            inds = best_sets[Ldim][:, 0] != np.inf
            nptest.assert_array_less(best_sets[Ldim][inds, 0],
                                     self.measskew_tol)

    def test_chooseOptQoIs_large_verbose(self):
        """
        Test :meth:`bet.sensitivity.chooseQoIs.chooseOptQoIs_large_verbose`.
        """
        self.qoiIndices = np.arange(0, self.output_dim)
        [best_sets, optsingvals_list] = cQoIs.chooseOptQoIs_large_verbose(
            self.input_set_centers, qoiIndices=self.qoiIndices,
            num_optsets_return=self.num_optsets_return,
            inner_prod_tol=self.inner_prod_tol, measskew_tol=self.measskew_tol)

        # Test that input_dim - 1 optsingval tensors are returned
        self.assertEqual(len(optsingvals_list), self.input_dim - 1)

        # Test that each tensor is the right shape
        for i in range(self.input_dim - 1):
            self.assertEqual(optsingvals_list[i].shape, (self.num_centers,
                                                         i + 2, self.num_optsets_return))


class test_2to20_choose2(ChooseQoIsMethods, unittest.TestCase):
    def setUp(self):
        self.input_dim = 2
        self.input_set = sample.sample_set(self.input_dim)
        self.input_set_centers = sample.sample_set(self.input_dim)
        self.output_dim_return = 2
        self.num_optsets_return = 5
        self.radius = 0.01
        np.random.seed(0)
        self.num_centers = 10
        self.centers = np.random.random((self.num_centers, self.input_dim))
        self.input_set_centers.set_values(self.centers)
        self.input_set = grad.sample_l1_ball(self.input_set_centers,
                                             self.input_dim + 1, self.radius)

        self.output_dim = 20
        self.output_set = sample.sample_set(self.output_dim)
        coeffs = np.random.random((self.input_dim,
                                   self.output_dim - self.input_dim))
        self.coeffs = np.append(coeffs, np.eye(self.input_dim), axis=1)

        self.output_set.set_values(self.input_set._values.dot(self.coeffs))
        self.my_disc = sample.discretization(self.input_set,
                                             self.output_set)
        self.center_disc = grad.calculate_gradients_rbf(
            self.my_disc, self.num_centers)
        self.input_set_centers = self.center_disc.get_input_sample_set()
        self.inner_prod_tol = 1.0
        self.measskew_tol = 100.0


class test_4to20_choose4(ChooseQoIsMethods, unittest.TestCase):
    def setUp(self):
        self.input_dim = 4
        self.input_set = sample.sample_set(self.input_dim)
        self.input_set_centers = sample.sample_set(self.input_dim)
        self.output_dim_return = 4
        self.num_optsets_return = 5
        self.radius = 0.01
        np.random.seed(0)
        self.num_centers = 100
        self.centers = np.random.random((self.num_centers, self.input_dim))
        self.input_set_centers.set_values(self.centers)
        self.input_set = grad.sample_l1_ball(self.input_set_centers,
                                             self.input_dim + 1, self.radius)

        self.output_dim = 20
        self.output_set = sample.sample_set(self.output_dim)
        coeffs = np.random.random((self.input_dim,
                                   self.output_dim - self.input_dim))
        self.coeffs = np.append(coeffs, np.eye(self.input_dim), axis=1)

        self.output_set.set_values(self.input_set._values.dot(self.coeffs))
        self.my_disc = sample.discretization(self.input_set,
                                             self.output_set)
        self.center_disc = grad.calculate_gradients_rbf(
            self.my_disc, self.num_centers)
        self.input_set_centers = self.center_disc.get_input_sample_set()

        self.inner_prod_tol = 0.9
        self.measskew_tol = 20.0


class test_9to15_choose9(ChooseQoIsMethods, unittest.TestCase):
    def setUp(self):
        self.input_dim = 9
        self.input_set = sample.sample_set(self.input_dim)
        self.input_set_centers = sample.sample_set(self.input_dim)
        self.output_dim_return = 9
        self.num_optsets_return = 50
        self.radius = 0.01
        np.random.seed(0)
        self.num_centers = 15
        self.centers = np.random.random((self.num_centers, self.input_dim))
        self.input_set_centers.set_values(self.centers)
        self.input_set = grad.sample_l1_ball(self.input_set_centers,
                                             self.input_dim + 1, self.radius)

        self.output_dim = 15
        self.output_set = sample.sample_set(self.output_dim)
        coeffs = np.random.random((self.input_dim,
                                   self.output_dim - self.input_dim))
        self.coeffs = np.append(coeffs, np.eye(self.input_dim), axis=1)

        self.output_set.set_values(self.input_set._values.dot(self.coeffs))
        self.my_disc = sample.discretization(self.input_set,
                                             self.output_set)
        self.center_disc = grad.calculate_gradients_rbf(
            self.my_disc, self.num_centers)
        self.input_set_centers = self.center_disc.get_input_sample_set()

        self.inner_prod_tol = 0.8
        self.measskew_tol = 100.0


class test_9to15_choose4(ChooseQoIsMethods, unittest.TestCase):
    def setUp(self):
        self.input_dim = 9
        self.input_set = sample.sample_set(self.input_dim)
        self.input_set_centers = sample.sample_set(self.input_dim)
        self.output_dim_return = 4
        self.num_optsets_return = 1
        self.radius = 0.01
        np.random.seed(0)
        self.num_centers = 11
        self.centers = np.random.random((self.num_centers, self.input_dim))
        self.input_set_centers.set_values(self.centers)
        self.input_set = grad.sample_l1_ball(self.input_set_centers,
                                             self.input_dim + 1, self.radius)

        self.output_dim = 15
        self.output_set = sample.sample_set(self.output_dim)
        coeffs = np.random.random((self.input_dim, self.output_dim -
                                   self.input_dim))
        self.coeffs = np.append(coeffs, np.eye(self.input_dim), axis=1)

        self.output_set.set_values(self.input_set._values.dot(self.coeffs))
        self.my_disc = sample.discretization(self.input_set,
                                             self.output_set)
        self.center_disc = grad.calculate_gradients_rbf(
            self.my_disc, self.num_centers)
        self.input_set_centers = self.center_disc.get_input_sample_set()

        self.inner_prod_tol = 0.9
        self.measskew_tol = 50.0


class test_2to28_choose2_zeros(ChooseQoIsMethods, unittest.TestCase):
    def setUp(self):
        self.input_dim = 2
        self.input_set = sample.sample_set(self.input_dim)
        self.input_set_centers = sample.sample_set(self.input_dim)
        self.output_dim_return = 2
        self.num_optsets_return = 5
        self.radius = 0.01
        np.random.seed(0)
        self.num_centers = 10
        self.centers = np.random.random((self.num_centers, self.input_dim))
        self.input_set_centers.set_values(self.centers)
        self.input_set = grad.sample_l1_ball(self.input_set_centers,
                                             self.input_dim + 1, self.radius)

        self.output_dim = 28
        self.output_set = sample.sample_set(self.output_dim)
        coeffs = np.ones((self.input_dim, 2 * self.input_dim))
        coeffs = np.append(coeffs, np.random.random((self.input_dim,
                                                     self.output_dim - 3 * self.input_dim)), axis=1)
        self.coeffs = np.append(coeffs, np.eye(self.input_dim), axis=1)

        self.output_set.set_values(self.input_set._values.dot(self.coeffs))
        self.my_disc = sample.discretization(self.input_set,
                                             self.output_set)
        self.center_disc = grad.calculate_gradients_rbf(
            self.my_disc, self.num_centers)
        self.input_set_centers = self.center_disc.get_input_sample_set()

        self.inner_prod_tol = 0.9
        self.measskew_tol = np.inf



# Copyright (C) 2014-2020 The BET Development Team


import unittest
import os
import glob
import numpy as np
import numpy.testing as nptest
import bet
import bet.sample as sample
import bet.util as util
import bet.surrogates as surrogates
import bet.sampling.basicSampling as bsam
import bet.calculateP.simpleFunP as simpleFunP
from bet.Comm import comm, MPI


def linear_model1(parameter_samples):
    Q_map = np.array([[0.506, 0.463], [0.253, 0.918], [0.085, 0.496]])
    QoI_samples = np.dot(parameter_samples, Q_map)
    return QoI_samples


def linear_model2(parameter_samples):
    Q_map = np.array([[0.506], [0.253], [0.085]])
    QoI_samples = np.dot(parameter_samples, Q_map)
    return QoI_samples


def linear_model3(parameter_samples):
    Q_map = np.array([[0.506]])
    QoI_samples = np.dot(parameter_samples, Q_map)
    return QoI_samples


class Test_piecewise_polynomial_surrogate_3_to_2(unittest.TestCase):
    """
    Testing :meth:`bet.surrogates.piecewise_polynomial_surrogate` on a
    3 to 2 map.

    """

    def setUp(self):
        """
        Setup map.
        """
        param_ref = np.array([0.5, 0.5, 0.5])
        Q_ref = linear_model1(param_ref)

        sampler = bsam.sampler(linear_model1)
        input_samples = sample.sample_set(3)
        input_samples.set_domain(np.repeat([[0.0, 1.0]], 3, axis=0))
        input_samples = sampler.random_sample_set(
            'uniform', input_samples, num_samples=1E2)
        disc = sampler.compute_qoi_and_create_discretization(input_samples,
                                                             globalize=True)
        simpleFunP.regular_partition_uniform_distribution_rectangle_scaled(
            data_set=disc, Q_ref=Q_ref, rect_scale=0.5)
        num = disc.check_nums()
        disc._output_sample_set.set_error_estimates(0.01 * np.ones((num, 2)))
        jac = np.zeros((num, 2, 3))
        jac[:, :, :] = np.array(
            [[0.506, 0.463], [0.253, 0.918], [0.085, 0.496]]).transpose()

        disc._input_sample_set.set_jacobians(jac)
        self.sur = surrogates.piecewise_polynomial_surrogate(disc)

    def test_constants(self):
        """
        Test for piecewise constants.
        """
        iss = bsam.random_sample_set('uniform',
                                     self.sur.input_disc._input_sample_set.get_dim(),
                                     num_samples=30,
                                     globalize=False)
        sur_disc = self.sur.generate_for_input_set(iss, order=0)
        sur_disc.check_nums()
        self.assertEqual(sur_disc._output_sample_set._dim, 2)
        nptest.assert_array_equal(sur_disc._input_sample_set._domain,
                                  self.sur.input_disc._input_sample_set._domain)
        sur_disc._input_sample_set._values_local[0, :]
        s_set = sur_disc._input_sample_set.copy()
        sur_disc.set_io_ptr()
        regions_local = np.equal(sur_disc._io_ptr_local, 0)
        s_set.set_region_local(regions_local)
        s_set.local_to_global()
        s_set.check_num()
        self.sur.calculate_prob_for_sample_set_region(s_set,
                                                      regions=[0],
                                                      update_input=True)

    def test_linears(self):
        """
        Test for piecewise linears.
        """
        iss = bsam.random_sample_set('uniform',
                                     self.sur.input_disc._input_sample_set.get_dim(),
                                     num_samples=20,
                                     globalize=False)
        sur_disc = self.sur.generate_for_input_set(iss, order=1)
        sur_disc.check_nums()
        self.assertEqual(sur_disc._output_sample_set._dim, 2)
        nptest.assert_array_equal(sur_disc._input_sample_set._domain,
                                  self.sur.input_disc._input_sample_set._domain)
        sur_disc._input_sample_set._values_local[0, :]
        s_set = sur_disc._input_sample_set.copy()
        sur_disc.set_io_ptr()
        regions_local = np.equal(sur_disc._io_ptr_local, 0)
        s_set.set_region_local(regions_local)
        s_set.local_to_global()
        s_set.check_num()
        self.sur.calculate_prob_for_sample_set_region(s_set,
                                                      regions=[0],
                                                      update_input=True)


class Test_piecewise_polynomial_surrogate_3_to_1(unittest.TestCase):
    """
    Testing :meth:`bet.surrogates.piecewise_polynomial_surrogate` on a
    3 to 1 map.

    """

    def setUp(self):
        """
        Setup problem.
        """
        param_ref = np.array([0.5, 0.5, 0.5])
        Q_ref = linear_model2(param_ref)

        sampler = bsam.sampler(linear_model2)
        input_samples = sample.sample_set(3)
        input_samples.set_domain(np.repeat([[0.0, 1.0]], 3, axis=0))
        input_samples = sampler.random_sample_set(
            'uniform', input_samples, num_samples=1E2)
        disc = sampler.compute_qoi_and_create_discretization(input_samples,
                                                             globalize=True)
        simpleFunP.regular_partition_uniform_distribution_rectangle_scaled(
            data_set=disc, Q_ref=Q_ref, rect_scale=0.5)
        num = disc.check_nums()
        disc._output_sample_set.set_error_estimates(0.01 * np.ones((num, 1)))
        jac = np.zeros((num, 1, 3))
        jac[:, :, :] = np.array([[0.506], [0.253], [0.085]]).transpose()

        disc._input_sample_set.set_jacobians(jac)
        self.sur = surrogates.piecewise_polynomial_surrogate(disc)

    def test_constants(self):
        """
        Test for piecewise constants.
        """
        iss = bsam.random_sample_set('uniform',
                                     self.sur.input_disc._input_sample_set.get_dim(),
                                     num_samples=30,
                                     globalize=False)
        sur_disc = self.sur.generate_for_input_set(iss, order=0)
        sur_disc.check_nums()
        self.assertEqual(sur_disc._output_sample_set._dim, 1)
        nptest.assert_array_equal(sur_disc._input_sample_set._domain,
                                  self.sur.input_disc._input_sample_set._domain)
        sur_disc._input_sample_set._values_local[0, :]

        s_set = sur_disc._input_sample_set.copy()
        sur_disc.set_io_ptr()
        regions_local = np.equal(sur_disc._io_ptr_local, 0)
        s_set.set_region_local(regions_local)
        s_set.local_to_global()
        s_set.check_num()
        self.sur.calculate_prob_for_sample_set_region(s_set,
                                                      regions=[0],
                                                      update_input=True)

    def test_linears(self):
        """
        Test for piecewise linears.
        """
        iss = bsam.random_sample_set('uniform',
                                     self.sur.input_disc._input_sample_set.get_dim(),
                                     num_samples=20,
                                     globalize=False)
        sur_disc = self.sur.generate_for_input_set(iss, order=1)
        sur_disc.check_nums()
        self.assertEqual(sur_disc._output_sample_set._dim, 1)
        nptest.assert_array_equal(sur_disc._input_sample_set._domain,
                                  self.sur.input_disc._input_sample_set._domain)
        sur_disc._input_sample_set._values_local[0, :]

        s_set = sur_disc._input_sample_set.copy()
        sur_disc.set_io_ptr()
        regions_local = np.equal(sur_disc._io_ptr_local, 0)
        s_set.set_region_local(regions_local)
        s_set.local_to_global()
        s_set.check_num()
        self.sur.calculate_prob_for_sample_set_region(s_set,
                                                      regions=[0],
                                                      update_input=True)


class Test_piecewise_polynomial_surrogate_1_to_1(unittest.TestCase):
    """
    Testing :meth:`bet.surrogates.piecewise_polynomial_surrogate` on a
    1 to 1 map.

    """

    def setUp(self):
        """
        Setup maps
        """
        param_ref = np.array([0.5])
        Q_ref = linear_model3(param_ref)

        sampler = bsam.sampler(linear_model3)
        input_samples = sample.sample_set(1)
        input_samples.set_domain(np.repeat([[0.0, 1.0]], 1, axis=0))
        input_samples = sampler.random_sample_set(
            'uniform', input_samples, num_samples=1E3)
        disc = sampler.compute_qoi_and_create_discretization(input_samples,
                                                             globalize=True)
        simpleFunP.regular_partition_uniform_distribution_rectangle_scaled(
            data_set=disc, Q_ref=Q_ref, rect_scale=0.5)
        num = disc.check_nums()
        disc._output_sample_set.set_error_estimates(0.01 * np.ones((num, 1)))
        jac = np.zeros((num, 1, 1))
        jac[:, :, :] = np.array([[0.506]]).transpose()

        disc._input_sample_set.set_jacobians(jac)
        self.sur = surrogates.piecewise_polynomial_surrogate(disc)

    def test_constants(self):
        """
        Test methods for order 0 polynomials.
        """
        iss = bsam.random_sample_set('uniform',
                                     self.sur.input_disc._input_sample_set.get_dim(),
                                     num_samples=20,
                                     globalize=False)
        sur_disc = self.sur.generate_for_input_set(iss, order=0)
        sur_disc.check_nums()
        self.assertEqual(sur_disc._output_sample_set._dim, 1)
        nptest.assert_array_equal(sur_disc._input_sample_set._domain,
                                  self.sur.input_disc._input_sample_set._domain)
        sur_disc._input_sample_set._values_local[0, :]

        s_set = sur_disc._input_sample_set.copy()
        sur_disc.set_io_ptr()
        regions_local = np.equal(sur_disc._io_ptr_local, 0)
        s_set.set_region_local(regions_local)
        s_set.local_to_global()
        s_set.check_num()
        self.sur.calculate_prob_for_sample_set_region(s_set,
                                                      regions=[0],
                                                      update_input=True)

    def test_linears(self):
        """
        Test for piecewise linears.
        """
        iss = bsam.random_sample_set('uniform',
                                     self.sur.input_disc._input_sample_set.get_dim(),
                                     num_samples=20,
                                     globalize=False)
        sur_disc = self.sur.generate_for_input_set(iss, order=1)
        sur_disc.check_nums()
        self.assertEqual(sur_disc._output_sample_set._dim, 1)
        nptest.assert_array_equal(sur_disc._input_sample_set._domain,
                                  self.sur.input_disc._input_sample_set._domain)
        sur_disc._input_sample_set._values_local[0, :]

        s_set = sur_disc._input_sample_set.copy()
        sur_disc.set_io_ptr()
        regions_local = np.equal(sur_disc._io_ptr_local, 0)
        s_set.set_region_local(regions_local)
        s_set.local_to_global()
        s_set.check_num()
        self.sur.calculate_prob_for_sample_set_region(s_set,
                                                      regions=[0],
                                                      update_input=True)



# Copyright (C) 2014-2020 The BET Development Team

import bet.sampling.basicSampling as bsam
import bet.calculateP.calculateR as calculateR
import bet.sampling.useLUQ as useLUQ
import bet.postProcess.plotP as plotP
import bet.postProcess.compareP as compP
import numpy as np

"""
Use LUQ to solve the Sel'kov model for glycolysis and learn quantities of interest.
This also illustrates several different options available within `calculateR`  to approximate the updated density.

The LUQ package must be installed to run this example.
"""

# sample for prediction set
p_set = bsam.random_sample_set(rv=[['uniform', {'loc': .01, 'scale': 0.114}],
                              ['uniform', {'loc': .05, 'scale': 1.45}]],
                               input_obj=2, num_samples=500)

# sample for observation set
o_set = bsam.random_sample_set(rv=[['beta', {'a': 2, 'b': 2, 'loc': .01, 'scale': 0.114}],
                                   ['beta', {'a': 2, 'b': 2, 'loc': .05, 'scale': 1.45}]],
                               input_obj=2, num_samples=500)

# Construct the predicted time series data
time_start = 2.0
time_end = 6.5
num_time_preds = int((time_end-time_start)*100)  # number of predictions (uniformly space) between [time_start,time_end]
times = np.linspace(time_start, time_end, num_time_preds)

# Initialize and setup LUQ
luq = useLUQ.useLUQ(predict_set=p_set, obs_set=o_set, lb_model=useLUQ.myModel, times=times)
luq.setup()

# Clean the data
time_start_idx = 0
time_end_idx = len(luq.times) - 1
luq.clean_data(time_start_idx=time_start_idx, time_end_idx=time_end_idx,
               num_clean_obs=20, tol=5.0e-2, min_knots=3, max_knots=12)

# Cluster and classify the dynamics
luq.dynamics(cluster_method='kmeans', kwargs={'n_clusters': 3, 'n_init': 10})

# Learn quantities of interest and transform the data
luq.learn_qois_and_transform(num_qoi=2)

# Convert LUQ output to discretization objects
disc1, disc2 = luq.make_disc()

# Set labels
param_labels = [r'$a$', r'$b$']

# Calculate initial total variation
comp_init = compP.compare(disc1, disc2, set1_init=True, set2_init=True)
print("Initial TV")
for i in range(2):
    print(comp_init.distance_marginal_quad(i=i, compare_factor=0.2, rtol=1.0e-3, maxiter=100))
# Invert to multivariate Gaussian
print("------------------------------------------------------")
print("Multivariate Gaussian")
calculateR.invert_to_multivariate_gaussian(disc1)

# Plot marginal probabilities and calculate total variations between probability measures
for i in range(2):
    plotP.plot_1d_marginal_densities(sets=(disc1.get_input_sample_set(), disc2.get_input_sample_set()), i=i,
                        sets_label_initial=['Initial', 'Data-Generating'], sets_label=['Updated', ''],
                        title="Multivariate Gaussian", label=param_labels[i])

# Calculate updated total variation
comp_init = compP.compare(disc1, disc2, set1_init=False, set2_init=True)
print("Updated TV")
for i in range(2):
    print(comp_init.distance_marginal_quad(i=i, compare_factor=0.2, rtol=1.0e-3, maxiter=100))

# Invert to Gaussian Mixture Model
print("------------------------------------------------------")
print("Gaussian Mixture Model")
calculateR.invert_to_gmm(disc1)
for i in range(2):
    plotP.plot_1d_marginal_densities(sets=(disc1.get_input_sample_set(), disc2.get_input_sample_set()), i=i,
                        sets_label_initial=['Initial', 'Data-Generating'], sets_label=['Updated', ''],
                        title="Gaussian Mixture Model", label=param_labels[i])
# Calculate updated total variation
comp_init = compP.compare(disc1, disc2, set1_init=False, set2_init=True)
print("Updated TV")
for i in range(2):
    print(comp_init.distance_marginal_quad(i=i, compare_factor=0.2, rtol=1.0e-3, maxiter=100))

print("------------------------------------------------------")
print("Weighted Kernel Density Estimate")
calculateR.invert_to_kde(disc1)
for i in range(2):
    plotP.plot_1d_marginal_densities(sets=(disc1.get_input_sample_set(), disc2.get_input_sample_set()), i=i,
                        sets_label_initial=['Initial', 'Data-Generating'], sets_label=['Updated', ''],
                        title="Weighted KDEs", label=param_labels[i]
                        )
# Calculate updated total variation
comp_init = compP.compare(disc1, disc2, set1_init=False, set2_init=True)
print("Updated TV")
for i in range(2):
    print(comp_init.distance_marginal_quad(i=i, compare_factor=0.2, rtol=1.0e-3, maxiter=100))

print("------------------------------------------------------")
print("Beta distribution")

calculateR.invert_to_random_variable(disc1, rv='beta')
for i in range(2):
    plotP.plot_1d_marginal_densities(sets=(disc1.get_input_sample_set(), disc2.get_input_sample_set()), i=i,
                        sets_label_initial=['Initial', 'Data-Generating'], sets_label=['Updated', ''],
                        title="Fitted Beta Distribution", label=param_labels[i]
                        )
# Calculate updated total variation
comp_init = compP.compare(disc1, disc2, set1_init=False, set2_init=True)
print("Updated TV")
for i in range(2):
    print(comp_init.distance_marginal_quad(i=i, compare_factor=0.2, rtol=1.0e-3, maxiter=100))



# Copyright (C) 2014-2020 The BET Development Team
"""
The module contains a class for interfacing between BET and LUQ.
"""

import numpy as np
import bet.sample as sample
import bet.util as util
import logging

class missing_module(Exception):
    """
    Exception for when a module cannot be imported.
    """

def myModel(inputs, times):
    """
    Example for interfacing a time series model with LUQ.
    :param inputs: Parameter values at which to evaluate the model.
    :type inputs: :class:`numpy.ndarray` of shape (num_inputs, num_params)
    :param times: Times at which to output results.
    :type times: :class:`numpy.ndarray` of shape (num_times, )
    :return: Time series data
    :rtype: :class:`numpy.ndarray` of shape (num_inputs, num_times)
    """
    try:
        from luq.dynamical_systems import Selkov
    except ImportError:
        raise missing_module("luq cannot be imported")
    ics = np.ones(inputs.shape)
    # Solve systems
    phys = Selkov()
    return phys.solve(ics=ics, params=inputs, t_eval=times)


class useLUQ:
    """
    Wrappers for interfacing BET with LUQ. Allows for the simple creation of `bet.sample.discretization` objects
    from LUQ output.
    """

    def __init__(self, predict_set, lb_model, times, obs_set=None):
        """
        Initialize the object.
        :param predict_set: Sample set defining input prediction samples.
        :type predict_set: :class:`bet.sample.sample_set`
        :param obs_set: Sample set defining input observation samples.
        :type obs_set: :class:`bet.sample.sample_set`
        :param lb_model: Interface to a time-dependent model takes an input of an array of parameter values and an array
        of times for evaluation as arguments. See an example with `myModel`, above.
        :param times: Times at which to output the model.
        :type times: :class:`numpy.ndarray` with shape (num_times, )
        """

        self.predict_set = predict_set
        self.obs_set = obs_set
        self.lb_model = lb_model
        self.times = times
        self.predicted_time_series = None
        self.obs_time_series = None
        self.learn = None

    def save(self, savefile):
        """
        Save the object to a Pickle file.
        :param savefile: Name of file to save to.
        :type savefile: str
        """
        util.save_object(save_set=self, file_name=savefile, globalize=True)

    def get_predictions(self):
        """
        Evaluate the model for the predicted time series.
        """
        self.predicted_time_series = self.lb_model(self.predict_set.get_values(), self.times)

    def get_obs(self):
        """
        Evaluate the model for the predicted time series.
        """
        self.obs_time_series = self.lb_model(self.obs_set.get_values(), self.times)

    def set_observed_time_series(self, obs_time_series):
        """
        Set observed time series data manually.
        :param obs_time_series: time series data
        :type obs_time_series:
        :return: :class:`numpy.ndarray` with shape (num_obs, num_times)
        """
        self.obs_time_series = obs_time_series

    def initialize(self, predicted_time_series=None, obs_time_series=None, times=None):
        """
        Initialize the LUQ object. This can be used manually if time series are pre-computed.

        :param predicted_time_series: Time series solutions for predicted values.
        :type predicted_time_series: :class:`numpy.ndarray` of shape (num_predict_samples, num_times)
        :param obs_time_series: Time series solutions for predicted values.
        :type obs_time_series: :class:`numpy.ndarray` of shape (num_obs_samples, num_times)
        :param times: Times at which the series are output.
        :type times: :class:`numpy.ndarray` with shape (num_times, )
        """
        try:
            from luq.luq import LUQ
        except ImportError:
            raise missing_module("luq cannot be imported")

        if predicted_time_series is None:
            predicted_time_series = self.predicted_time_series
        if obs_time_series is None:
            obs_time_series = self.obs_time_series
        if times is None:
            times = self.times

        self.learn = LUQ(predicted_time_series, obs_time_series, times)

    def setup(self):
        """
        Setup LUQ object all at once.
        """
        self.get_predictions()
        self.get_obs()
        self.initialize(self.predicted_time_series, self.obs_time_series, self.times)

    def clean_data(self, **kwargs):
        """
        Wrapper for `luq.luq.LUQ.clean_data`
        """
        self.learn.clean_data(**kwargs)

    def dynamics(self, **kwargs):
        """
        Wrapper for `luq.luq.LUQ.dynamics`
        """
        self.learn.dynamics(**kwargs)

    def learn_qois_and_transform(self, **kwargs):
        """
        Wrapper for `luq.luq.LUQ.learn_qois_and_transform`
        """
        self.learn.learn_qois_and_transform(**kwargs)

    def make_disc(self):
        """
        Construct `bet.sample.discretization` objects for predict and obs sets.
        :return: predict_disc, obs_disc
        :rtype: `bet.sample.discretization`, `bet.sample.discretization` or None if no observation set.
        """
        out_dim = self.learn.num_pcs[0]

        predict_output = sample.sample_set(out_dim)
        predict_output.set_region_local(self.learn.predict_labels)
        predict_output.set_cluster_maps(self.learn.predict_maps)

        obs_output = sample.sample_set(out_dim)
        obs_output.set_region_local(self.learn.obs_labels)
        obs_output.set_cluster_maps(self.learn.obs_maps)

        # Prediction discretization
        disc1 = sample.discretization(input_sample_set=self.predict_set,
                                      output_sample_set=predict_output,
                                      output_observed_set=obs_output)
        disc1.local_to_global()

        # Observation discretization
        if self.obs_set is None:
            disc2 = None
        else:
            disc2 = sample.discretization(input_sample_set=self.obs_set,
                                          output_sample_set=obs_output)
            disc2.local_to_global()

        return disc1, disc2

    def local_to_global(self):
        """
        Dummy function for saving.
        """
        pass










#! /usr/bin/env python

# Copyright (C) 2014-2020 The BET Development Team

# import necessary modules
import numpy as np
import bet.postProcess.plotDomains as pDom
import scipy.io as sio
import bet.sample as sample

# Set minima and maxima
lam_domain = np.array([[.07, .15], [.1, .2]])

# Select only the stations I care about this will lead to better sampling
station_nums = [0, 5]  # 1, 6

# Read in Q_ref and Q to create the appropriate rho_D
mdat = sio.loadmat('Q_2D.mat')
Q = mdat['Q']
Q = Q[:, station_nums]
Q_ref = mdat['Q_true']
Q_ref = Q_ref[15, station_nums]  # 16th/20
bin_ratio = 0.15
bin_size = (np.max(Q, 0) - np.min(Q, 0)) * bin_ratio

# Create kernel
maximum = 1 / np.product(bin_size)


def rho_D(outputs):
    rho_left = np.repeat([Q_ref - .5 * bin_size], outputs.shape[0], 0)
    rho_right = np.repeat([Q_ref + .5 * bin_size], outputs.shape[0], 0)
    rho_left = np.all(np.greater_equal(outputs, rho_left), axis=1)
    rho_right = np.all(np.less_equal(outputs, rho_right), axis=1)
    inside = np.logical_and(rho_left, rho_right)
    max_values = np.repeat(maximum, outputs.shape[0], 0)
    return inside.astype('float64') * max_values


# Read in points_ref and plot results
ref_sample = mdat['points_true']
ref_sample = ref_sample[5:7, 15]

# Create input, output, and discretization from data read from file
points = mdat['points']
input_sample_set = sample.sample_set(points.shape[0])
input_sample_set.set_values(points.transpose())
input_sample_set.set_domain(lam_domain)
output_sample_set = sample.sample_set(Q.shape[1])
output_sample_set.set_values(Q)
my_disc = sample.discretization(input_sample_set, output_sample_set)

# Show the samples in the parameter space
pDom.scatter_rhoD(my_disc, rho_D=rho_D, ref_sample=ref_sample, io_flag='input')
# Show the corresponding samples in the data space
pDom.scatter_rhoD(output_sample_set, rho_D=rho_D, ref_sample=Q_ref,
                  io_flag='output')
# Show the data domain that corresponds with the convex hull of samples in the
# parameter space
pDom.show_data_domain_2D(my_disc, Q_ref=Q_ref)

# Show multiple data domains that correspond with the convex hull of samples in
# the parameter space
pDom.show_data_domain_multi(my_disc, Q_ref=mdat['Q_true'][15],
                            showdim='all')



# Copyright (C) 2014-2020 The BET Development Team

"""
This package contains all of the tests for :program:`BET`. The package
structure mirrors the ``bet`` package structure.
"""
__all__ = ['test_calculateP', 'test_postProcess', 'test_sampling',
           'test_sensitivity', 'test_util', 'test_Comm', 'test_sample',
           'test_surrogates', 'problem_setups']
